import os
import warnings

import numpy as np
import h5py
import so3g
from typing import Optional
from pixell import bunch, enmap, tilemap
from pixell import utils as putils

from . import utils as smutils
from .. import coords
from .pointing_matrix import PmatCut
from .noise_model import NmatUncorr

class MLMapmaker:
    def __init__(
        self,
        signals=[],
        noise_model=None,
        dtype=np.float32,
        verbose=False,
        glitch_flags: str = "flags.glitch_flags",
    ):
        """Initialize a Maximum Likelihood Mapmaker.
        Arguments:
        * signals: List of Signal-objects representing the models that will be solved
          jointly for. Typically this would be the sky map and the cut samples. NB!
          The way the cuts currently work, they *MUST* be the first signal specified.
          If not, the equation system will be inconsistent and won't converge.
        * noise_model: A noise model constructor which will be used to initialize the
          noise model for each observation. Can be overriden in add_obs.
        * dtype: The data type to use for the time-ordered data. Only tested with float32
        * verbose: Whether to print progress messages. Not implemented"""
        if noise_model is None:
            noise_model = NmatUncorr()
        self.signals = signals
        self.dtype = dtype
        self.verbose = verbose
        self.noise_model = noise_model
        self.data = []
        self.dof = smutils.MultiZipper()
        self.ready = False
        self.glitch_flags_path = glitch_flags

    def add_obs(self, id, obs, deslope=True, noise_model=None, signal_estimate=None):
        # Prepare our tod
        ctime = obs.timestamps
        srate = (len(ctime) - 1) / (ctime[-1] - ctime[0])
        tod = obs.signal.astype(self.dtype, copy=False)
        # Subtract an existing estimate of the signal before estimating
        # the noise model, if available
        if signal_estimate is not None:
            tod -= signal_estimate
        if deslope:
            putils.deslope(tod, w=5, inplace=True)
        # Allow the user to override the noise model on a per-obs level
        if noise_model is None:
            noise_model = self.noise_model
        # Build the noise model from the obs unless a fully
        # initialized noise model was passed
        if noise_model.ready:
            nmat = noise_model
        else:
            try:
                nmat = noise_model.build(tod, srate=srate)
            except Exception as e:
                msg = f"FAILED to build a noise model for observation='{id}' : '{e}'"
                raise RuntimeError(msg)
        # Add back signal estimate
        if signal_estimate is not None:
            tod += signal_estimate
            # The signal estimate might not be desloped, so
            # adding it back can reintroduce a slope. Fix that here.
            if deslope:
                putils.deslope(tod, w=5, inplace=True)
        # And apply it to the tod
        tod = nmat.apply(tod)
        # Add the observation to each of our signals
        for signal in self.signals:
            signal.add_obs(id, obs, nmat, tod)
        # Save what we need about this observation
        self.data.append(
            bunch.Bunch(
                id=id,
                ndet=obs.dets.count,
                nsamp=len(ctime),
                dets=obs.dets.vals,
                nmat=nmat,
            )
        )

    def prepare(self):
        if self.ready:
            return
        for signal in self.signals:
            signal.prepare()
            self.dof.add(signal.dof)
        self.ready = True

    def evaluator(self, x_zip):
        """Return a helper object that lets one evaluate the data model
        Px for the zipped solution x for a TOD"""
        return MLEvaluator(x_zip, self.signals, self.dof, dtype=self.dtype)
    def accumulator(self):
        """Return a helper object that lets one accumulate P'd for
        for a TOD"""
        return MLAccumulator(self.signals, self.dof)

    def A(self, x_zip):
        # unzip goes from flat array of all the degrees of freedom to individual maps, cuts etc.
        # to_work makes a scratch copy and does any redistribution needed
        evaluator   = self.evaluator(x_zip)
        accumulator = self.accumulator()
        for di, data in enumerate(self.data):
            tod = evaluator.evaluate(data)
            data.nmat.apply(tod)
            accumulator.accumulate(data, tod)
        accumulator.finish()
        self.prior(evaluator.x, accumulator.x)
        return self.dof.zip(*accumulator.x)

    def M(self, x_zip):
        iwork = self.dof.unzip(x_zip)
        result = self.dof.zip(
            *[signal.precon(w) for signal, w in zip(self.signals, iwork)]
        )
        return result

    def solve(
        self,
        maxiter=500,
        maxerr=1e-9,
        x0=None,
        fname_checkpoint=None,
        checkpoint_interval=1,
    ):
        self.prepare()
        rhs = self.dof.zip(*[signal.rhs for signal in self.signals])

        solver = putils.CG(self.A, rhs, M=self.M, dot=self.dof.dot, x0=x0)
        # If there exists a checkpoint, restore solver state
        if fname_checkpoint is None:
            checkpoint = False
            restart = False
        else:
            checkpoint = True
            outdir = os.path.dirname(fname_checkpoint)
            if len(outdir) != 0:
                os.makedirs(outdir, exist_ok=True)
            if os.path.isfile(fname_checkpoint):
                solver.load(fname_checkpoint)
                restart = True
            else:
                restart = False
        while restart or (solver.i < maxiter and solver.err > maxerr):
            if restart:
                # When restarting, do not step
                restart = False
            else:
                solver.step()
                if checkpoint and solver.i % checkpoint_interval == 0:
                    # Avoid checkpoint corruption by making a copy of the previous checkpoint
                    if os.path.isfile(fname_checkpoint):
                        os.replace(fname_checkpoint, fname_checkpoint + ".old")
                    # Write a checkpoint
                    solver.save(fname_checkpoint)
            # x is the unzipped solution. It's a list of one object per signal we solve for.
            # x_zip is the raw solution, as a 1d vector.
            yield bunch.Bunch(i=solver.i, err=solver.err, x=self.dof.unzip(solver.x), x_zip=solver.x)

    def prior(self, xins, xouts):
        for signal, xin, xout in zip(self.signals, xins, xouts):
            signal.prior(xin, xout)

    def translate(self, other, x_zip):
        """Translate degrees of freedom x from some other mapamaker to the current one.
        The other mapmaker must have the same list of signals, except that they can have
        different sample rate etc. than this one. See the individual Signal-classes
        translate methods for details. This is used in multipass mapmaking.
        """
        x     = other.dof.unzip(x_zip)
        x_new = []
        for ssig, osig, oval in zip(self.signals, other.signals, x):
            x_new.append(ssig.translate(osig, oval))
        return self.dof.zip(*x_new)

class MLEvaluator:
    """Helper for MLMapmaker that represents the action of P in the model d = Px+n."""
    def __init__(self, x_zip, signals, dof, dtype=np.float32):
        self.signals = signals
        self.x_zip   = x_zip
        self.x       = dof.unzip(x_zip)
        self.dof     = dof
        self.dtype   = dtype
        self.iwork = [signal.to_work(m) for signal, m in zip(self.signals, self.x)]
    def evaluate(self, data, tod=None):
        """Evaluate Px for one tod"""
        if tod is None: tod = np.zeros([data.ndet, data.nsamp], self.dtype)
        for si, signal in reversed(list(enumerate(self.signals))):
            signal.forward(data.id, tod, self.iwork[si])
        return tod

class MLAccumulator:
    """Helper for MLMapmaker that represents the action of P.T in the model d = Px+n."""
    def __init__(self, signals, dof):
        self.signals = signals
        self.dof     = dof
        self.owork = [signal.wzeros() for signal in signals]
    def accumulate(self, data, tod):
        """Accumulate P'd for one tod"""
        for si, signal in enumerate(self.signals):
            signal.backward(data.id, tod, self.owork[si])
    def finish(self):
        """Return the full P'd based on the previous accumulation"""
        self.x = [signal.from_work(w) for signal, w in zip(self.signals, self.owork)]
        self.x_zip = self.dof.zip(*self.x)
        return self.x_zip

class Signal:
    """This class represents a thing we want to solve for, e.g. the sky, ground, cut samples, etc."""

    def __init__(self, name, ofmt, output, ext, **kwargs):
        """Initialize a Signal. It probably doesn't make sense to construct a generic signal
        directly, though. Use one of the subclasses.
        Arguments:
        * name: The name of this signal, e.g. "sky", "cut", etc.
        * ofmt: The format used when constructing output file prefix
        * output: Whether this signal should be part of the output or not.
        * ext: The extension used for the files.
        * **kwargs: additional keyword based parameters, accessible as class parameters
        """
        self.name = name
        self.ofmt = ofmt
        self.output = output
        self.ext = ext
        self.dof = None
        self.ready = False
        self.__dict__.update(kwargs)

    def add_obs(self, id, obs, nmat, Nd, **kwargs):
        pass

    def prepare(self):
        self.ready = True

    def forward(self, id, tod, x):
        pass

    def backward(self, id, tod, x):
        pass

    def precon(self, x):
        return x

    def to_work(self, x):
        return x.copy()

    def from_work(self, x):
        return x

    def wzeros(self):
        return 0

    def write(self, prefix, tag, x):
        pass

    def translate(self, other, x):
        return x

    def prior(self, xin, xout):
        pass


class SignalMap(Signal):
    """Signal describing a non-distributed sky map."""

    def __init__(
        self,
        shape,
        wcs,
        comm,
        comps="TQU",
        name="sky",
        ofmt="{name}",
        output=True,
        ext="fits",
        dtype=np.float32,
        sys=None,
        recenter=None,
        tile_shape=(500, 500),
        tiled=False,
        interpol=None,
        cut_field= "flags.glitch_flags",
    ):
        """Signal describing a sky map in the coordinate system given by "sys", which defaults
        to equatorial coordinates. If tiled==True, then this will be a distributed map with
        the given tile_shape, otherwise it will be a plain enmap. interpol controls the
        pointing matrix interpolation mode. See so3g's Projectionist docstring for details.
        """
        Signal.__init__(self, name, ofmt, output, ext)
        self.comm = comm
        self.comps = comps
        self.sys = sys
        self.recenter = recenter
        self.dtype = dtype
        self.tiled = tiled
        self.interpol = interpol
        self.data = {}
        self.cut_field = cut_field
        ncomp = len(comps)
        shape = tuple(shape[-2:])

        if tiled:
            geo = tilemap.geometry(shape, wcs, tile_shape=tile_shape)
            self.rhs = tilemap.zeros(geo.copy(pre=(ncomp,)), dtype=dtype)
            self.div = tilemap.zeros(geo.copy(pre=(ncomp, ncomp)), dtype=dtype)
            self.hits = tilemap.zeros(geo.copy(pre=()), dtype=dtype)
        else:
            self.rhs = enmap.zeros((ncomp,) + shape, wcs, dtype=dtype)
            self.div = enmap.zeros((ncomp, ncomp) + shape, wcs, dtype=dtype)
            self.hits = enmap.zeros(shape, wcs, dtype=dtype)

    def add_obs(self, id, obs, nmat, Nd, pmap=None, cuts=None):
        """Add and process an observation, building the pointing matrix
        and our part of the RHS. "obs" should be an Observation axis manager,
        nmat a noise model, representing the inverse noise covariance matrix,
        and Nd the result of applying the noise model to the detector time-ordered data.
        """
        Nd     = Nd.copy()  # This copy can be avoided if build_obs is split into two parts
        ctime  = obs.timestamps
        if cuts is None: cuts = obs[self.cut_field]
        pcut   = PmatCut(cuts)  # could pass this in, but fast to construct
        if pmap is None:
            # Build the local geometry and pointing matrix for this observation
            if self.recenter:
                rot = smutils.recentering_to_quat_lonlat(
                    *smutils.evaluate_recentering(
                        self.recenter,
                        ctime=ctime[len(ctime) // 2],
                        geom=self.rhs.geometry,
                        site=smutils.unarr(obs.site),
                    )
                )
            else:
                rot = None
            pmap = coords.pmat.P.for_tod(
                obs,
                comps=self.comps,
                geom=self.rhs.geometry,
                rot=rot,
                threads="domdir",
                weather=smutils.unarr(obs.weather),
                site=smutils.unarr(obs.site),
                interpol=self.interpol,
            )
        # Build the RHS for this observation
        pcut.clear(Nd)
        obs_rhs = pmap.zeros()
        pmap.to_map(dest=obs_rhs, signal=Nd)
        # Build the per-pixel inverse covmat for this observation
        obs_div = pmap.zeros(super_shape=(self.ncomp, self.ncomp))
        for i in range(self.ncomp):
            obs_div[i] = 0
            obs_div[i, i] = 1
            Nd[:] = 0
            pmap.from_map(obs_div[i], dest=Nd)
            pcut.clear(Nd)
            Nd = nmat.white(Nd)
            obs_div[i] = 0
            pmap.to_map(signal=Nd, dest=obs_div[i])
        # Build hitcount
        Nd[:] = 1
        pcut.clear(Nd)
        obs_hits = pmap.to_map(signal=Nd)
        del Nd

        # Update our full rhs and div. This works for both plain and distributed maps
        self.rhs = self.rhs.insert(obs_rhs, op=np.ndarray.__iadd__)
        self.div = self.div.insert(obs_div, op=np.ndarray.__iadd__)
        self.hits = self.hits.insert(obs_hits[0], op=np.ndarray.__iadd__)
        # Save the per-obs things we need. Just the pointing matrix in our case.
        # Nmat and other non-Signal-specific things are handled in the mapmaker itself.
        self.data[id] = bunch.Bunch(pmap=pmap, obs_geo=obs_rhs.geometry)

    def prepare(self):
        """Called when we're done adding everything. Sets up the map distribution,
        degrees of freedom and preconditioner."""
        if self.ready:
            return
        if self.tiled:
            self.geo_work = self.rhs.geometry
            self.rhs = tilemap.redistribute(self.rhs, self.comm)
            self.div = tilemap.redistribute(self.div, self.comm)
            self.hits = tilemap.redistribute(self.hits, self.comm)
            self.dof = smutils.TileMapZipper(
                self.rhs.geometry, dtype=self.dtype, comm=self.comm
            )
        else:
            if self.comm is not None:
                self.rhs = putils.allreduce(self.rhs, self.comm)
                self.div = putils.allreduce(self.div, self.comm)
                self.hits = putils.allreduce(self.hits, self.comm)
            self.dof = smutils.MapZipper(*self.rhs.geometry, dtype=self.dtype)
        self.idiv = smutils.safe_invert_div(self.div)
        self.ready = True

    @property
    def ncomp(self):
        return len(self.comps)

    def forward(self, id, tod, map, tmul=1, mmul=1):
        """map2tod operation. For tiled maps, the map should be in work distribution,
        as returned by unzip. Adds into tod."""
        if id not in self.data:
            return  # Should this really skip silently like this?
        if tmul != 1:
            tod *= tmul
        if mmul != 1:
            map = map * mmul
        self.data[id].pmap.from_map(dest=tod, signal_map=map, comps=self.comps)

    def backward(self, id, tod, map, tmul=1, mmul=1):
        """tod2map operation. For tiled maps, the map should be in work distribution,
        as returned by unzip. Adds into map"""
        if id not in self.data:
            return
        if tmul != 1:
            tod = tod * tmul
        if mmul != 1:
            map *= mmul
        self.data[id].pmap.to_map(signal=tod, dest=map, comps=self.comps)

    def precon(self, map):
        if self.tiled:
            return tilemap.map_mul(self.idiv, map)
        else:
            return enmap.map_mul(self.idiv, map)

    def to_work(self, map):
        if self.tiled:
            return tilemap.redistribute(map, self.comm, self.geo_work.active)
        else:
            return map.copy()

    def from_work(self, map):
        if self.tiled:
            return tilemap.redistribute(map, self.comm, self.rhs.geometry.active)
        else:
            if self.comm is None:
                return map
            else:
                return putils.allreduce(map, self.comm)

    def wzeros(self):
        """Like to_work, but zeroed instead of containing the signal. Much cheaper"""
        if self.tiled: return tilemap.zeros(self.geo_work, self.rhs.dtype)
        else:          return enmap.zeros(*self.rhs.geometry,  self.rhs.dtype)

    def write(self, prefix, tag, m, unit='K'):
        if not self.output:
            return
        oname = self.ofmt.format(name=self.name)
        oname = "%s%s_%s.%s" % (prefix, oname, tag, self.ext)
        if self.tiled:
            tilemap.write_map(oname, m, self.comm, extra={'BUNIT':unit})
        else:
            if self.comm is None or self.comm.rank == 0:
                enmap.write_map(oname, m, extra={'BUNIT':unit})
        return oname

    def _checkcompat(self, other):
        # Check if other is compatible with us. For SignalMap, we currently
        # only support direct equivalence
        if other.sys != self.sys or other.recenter != self.recenter:
            raise ValueError("Coordinate system mismatch")
        if other.tiled != self.tiled:
            raise ValueError("Tiling mismatch")
        if self.tiled:
            if other.rhs.geometry.shape != self.rhs.geometry.shape:
                raise ValueError("Geometry mismatch")
        else:
            if other.rhs.shape != self.rhs.shape:
                raise ValueError("Geometry mismatch")

    def translate(self, other, map):
        """Translate map from another SignalMap representation to the current,
        returning a new map. The new map may be a reference to the original."""
        # Currently we don't support any actual translation, but could handle
        # resolution changes in the future (probably not useful though)
        self._checkcompat(other)
        return map

class SignalCut(Signal):
    def __init__(
        self,
        comm,
        name="cut",
        ofmt="{name}_{rank:02}",
        dtype=np.float32,
        output=False,
        cut_type=None,
        cut_field = "flags.glitch_flags",
    ):
        """Signal for handling the ML solution for the values of the cut samples."""
        Signal.__init__(self, name, ofmt, output, ext="hdf")
        self.comm = comm
        self.data = {}
        self.dtype = dtype
        self.cut_type = cut_type
        self.off = 0
        self.rhs = []
        self.div = []
        self.cut_field = cut_field

    def add_obs(self, id, obs, nmat, Nd, cuts=None):
        """Add and process an observation. "obs" should be an Observation axis manager,
        nmat a noise model, representing the inverse noise covariance matrix,
        and Nd the result of applying the noise model to the detector time-ordered data.
        """
        Nd = Nd.copy()  # This copy can be avoided if build_obs is split into two parts
        if cuts is None: cuts = obs[self.cut_field]
        pcut = PmatCut(cuts, model=self.cut_type)
        # Build our RHS
        obs_rhs = np.zeros(pcut.njunk, self.dtype)
        pcut.backward(Nd, obs_rhs)
        # Build our per-pixel inverse covmat
        obs_div = np.ones(pcut.njunk, self.dtype)
        Nd[:] = 0
        pcut.forward(Nd, obs_div)
        Nd *= nmat.ivar[:, None]
        pcut.backward(Nd, obs_div)
        self.data[id] = bunch.Bunch(pcut=pcut, i1=self.off, i2=self.off + pcut.njunk)
        self.off += pcut.njunk
        self.rhs.append(obs_rhs)
        self.div.append(obs_div)

    def prepare(self):
        """Process the added observations, determining our degrees of freedom etc.
        Should be done before calling forward and backward."""
        if self.ready: return
        self.rhs = np.concatenate(self.rhs) if len(self.rhs) > 0 else np.zeros(0, self.dtype)
        self.div = np.concatenate(self.div) if len(self.div) > 0 else np.zeros(0, self.dtype)
        self.dof = smutils.ArrayZipper(self.rhs.shape, dtype=self.dtype, comm=self.comm)
        self.ready = True

    def forward(self, id, tod, junk):
        if id not in self.data: return
        d = self.data[id]
        d.pcut.forward(tod, junk[d.i1:d.i2])

    def precon(self, junk):
        return junk / self.div

    def backward(self, id, tod, junk):
        if id not in self.data: return
        d = self.data[id]
        d.pcut.backward(tod, junk[d.i1:d.i2])

    def wzeros(self): return np.zeros_like(self.rhs)

    def write(self, prefix, tag, m):
        if not self.output: return
        if self.comm is None:
            rank = 0
        else:
            rank = self.comm.rank
        oname = self.ofmt.format(name=self.name, rank=rank)
        oname = "%s%s_%s.%s" % (prefix, oname, tag, self.ext)
        with h5py.File(oname, "w") as hfile:
            hfile["data"] = m
        return oname

    def _checkcompat(self, other):
        if other.cut_type != self.cut_type:
            raise ValueError("Cut type mismatch")

    def translate(self, other, junk):
        """Translate junk degrees of freedom from one SignalCut representation
        to another, e.g. changing sample rate. Returns the result"""
        self._checkcompat(other)
        res = np.full(self.off, np.nan, self.dtype)
        for id in self.data:
            sdata = self.data[id]
            odata = other.data[id]
            so3g.translate_cuts(
                odata.pcut.cuts,
                sdata.pcut.cuts,
                sdata.pcut.model,
                sdata.pcut.params,
                junk[odata.i1:odata.i2],
                res [sdata.i1:sdata.i2],
            )
        # This check should be cheap enough to be worth it.
        # It's only called once every mapmaking pass anyway.
        # It should not trigger during normal circumstances,
        # but is useful for catching bugs when changing how
        # the cuts work.
        if np.any(~np.isfinite(res)):
            warnings.warn("Incompatible cuts in translate(): Probably caused by differing numbers of cut ranges. Skipping translate for these cuts")
            res[:] = 0
        return res

class SignalSrcsamp(SignalCut):
    def __init__(
        self,
        comm,
        mask,
        sys=None,
        recenter=None,
        name="srcsamp",
        ofmt="{name}_{rank:02}",
        dtype=np.float32,
        map_dtype=np.float64,
        output=False,
        # Prior parameters
        eps_edge=10, eps_core=0.01, redge=2*putils.arcmin,
    ):
        """Signal for eliminating model error near small areas in the map with bright objects"""
        SignalCut.__init__(self, comm, name=name, ofmt=ofmt, dtype=dtype, output=output, cut_type="full")
        self.map_dtype= map_dtype
        # Mask should have shape (1,ny,nx).
        # Non-tiled, so mask should be low-res! 4 arcmin should work fine.
        # That's just 25 MB for our patch
        mask          = mask.preflat[0]
        self.distmap  = mask.distance_transform().astype(map_dtype)
        self.mask     = mask.astype(map_dtype)
        self.sys      = sys
        self.recenter = recenter
        # Prior stuff. The prior breaks the degeneracy between the map
        # and these extra degrees offreedom. The prior strength changes
        # smoothly around the edge of the mask to prevent discontinuities
        # in the final map. This is controlled by eps_core, eps_edge and redge
        self.eps_edge = eps_edge
        self.eps_core = eps_core
        self.redge    = redge
        # Distance of of our extra degrees of freedom
        self.distsamps  = []

    def add_obs(self, id, obs, nmat, Nd):
        """Add and process an observation. "obs" should be an Observation axis manager,
        nmat a noise model, representing the inverse noise covariance matrix,
        and Nd the result of applying the noise model to the detector time-ordered data.
        """
        # First scan our mask to find which samples need this
        # treatment
        if self.recenter:
            rec = smutils.evaluate_recentering(self.recenter, ctime=ctime[len(ctime) // 2],
                    geom=(self.mask.shape, self.mask.wcs), site=smutils.unarr(obs.site))
            rot = smutils.recentering_to_quat_lonlat(*rec)
        else: rot = None
        pmap = coords.pmat.P.for_tod(obs, comps="T", geom=self.mask.geometry,
            rot=rot, threads="domdir", weather=smutils.unarr(obs.weather),
            site=smutils.unarr(obs.site), interpol="nearest")
        tod = np.zeros((obs.dets.count, obs.samps.count), Nd.dtype)
        pmap.from_map(self.mask, dest=tod)
        cuts= so3g.proj.RangesMatrix.from_mask(tod>0.5)
        # Then build our RHS as normal
        tod[:]  = Nd.copy()
        pcut    = PmatCut(cuts, model=self.cut_type)
        obs_rhs = np.zeros(pcut.njunk, self.dtype)
        pcut.backward(tod, obs_rhs)
        # Build our per-pixel inverse covmat
        # This assumes cut_type="full"!
        obs_div = np.zeros(pcut.njunk, self.dtype)
        tod[:]  = nmat.ivar[:,None]
        pcut.backward(tod, obs_div)
        # Get the distance of each of our samples from the mask
        # This assumes cut_type="full"!
        obs_dist= np.zeros(pcut.njunk, self.dtype)
        pmap.from_map(self.distmap, dest=tod)
        pcut.backward(tod, obs_dist)
        # Finished!
        self.data[id] = bunch.Bunch(pcut=pcut, i1=self.off, i2=self.off + pcut.njunk)
        self.off += pcut.njunk
        self.rhs.append(obs_rhs)
        self.div.append(obs_div)
        self.distsamps.append(obs_dist)

    def prepare(self):
        """Process the added observations, determining our degrees of freedom etc.
        Should be done before calling forward and backward."""
        if self.ready: return
        SignalCut.prepare(self)
        self.distsamps = np.concatenate(self.distsamps) if len(self.distsamps) > 0 else np.zeros(0, self.dtype)
        x             = np.minimum(self.distsamps/self.redge, 1)
        self.epsilon  = np.exp(np.log(self.eps_edge) * (1-x) + np.log(self.eps_core) * x)
        self.epsilon *= self.div
        # Can now delete distansamps to save memory. epsilon is the
        # thing we'll actually use in the prior
        self.distsamps= None

    def prior(self, xin, xout):
        xout += xin * self.epsilon

    def translate(self, other, junk):
        """Translate junk degrees of freedom from one SignalSrcsamp representation
        to another, e.g. changing sample rate. Currently simply initializes the
        new ones to zero, since so3g.translate_cuts can't handle it when the number
        of cut ranges per detector has changed, which can happen with the way we build
        the srcsamp cuts. It could be supported if necessary, but it's not high priority."""
        return np.zeros(self.off, self.dtype)

# Needed for multi-beam support in SignalMap. These should probably be moved somewhere
# eventually. Maybe Projectionist should be generalized to support multiple beams so
# shit somewhat fragile construction isn't needed in the first place.
class MultiProj(so3g.proj.Projectionist):
    def to_map(self, signal, assemblies, output=None, det_weights=None, threads=None, comps=None):
        for ai, assembly in enumerate(assemblies):
            output = super().to_map(signal, assembly, output=output, det_weights=det_weights,
                threads=threads[ai] if threads is not None else None, comps=comps)
        return output
    def to_weights(self, assemblies, output=None, det_weights=None, threads=None, comps=None):
        for ai, assembly in enumerate(assemblies):
            output = super().to_weights(signal, assembly, output=output, det_weights=det_weights,
                threads=threads[ai] if threads is not None else None, comps=comps)
        return output
    def from_map(self, src_map, assemblies, signal=None, comps=None):
        for ai, assembly in enumerate(assemblies):
            signal = super().from_map(src_map, assembly, signal=signal, comps=comps)
        return signal
    def get_active_tiles(self, assemblies, assign=False):
        infos = []
        for ai, assembly in enumerate(assemblies):
            infos.append(super().get_active_tiles(assembly, assign=assign))
        active = list(np.unique(np.concatenate([info["active_tiles"] for info in infos])))
        # Mapping from old active to new
        amap   = [putils.find(active, info["active_tiles"]) for info in infos]
        hits   = np.zeros(len(active),int)
        for i, info in enumerate(infos):
            hits[amap[i]] += info["hit_counts"]
        # group_ranges and group_tiles are harder. We skip them for now.
        # They are only used by the "tiles" thread partitioning, not the
        # standard "domdir"
        return {"active_tiles":active, "hit_counts":hits}
    def assign_threads(self, assemblies, method='domdir', n_threads=None):
        return [super().assign_threads(assembly, method=method, n_threads=n_threads) for assembly in assemblies]
    def assign_threads_from_map(self, assembly, tmap, n_threads=None):
        return [super().assign_threads_from_map(assembly, tmap, n_threads=n_threads) for assembly in assemblies]

class PmatMultibeam(coords.pmat.P):
    def __init__(self, fps, **kwargs):
        super().init(fp=fps[0], **kwargs)
        self.fps = fps
    @classmethod
    def for_tod(cls, tod, focal_planes, threads=None, **kwargs):
        tmp = super().for_tod(tod, **kwargs)
        fps = [coords.helpers.get_fplane(tod, focal_plane=fplane) for fplane in focal_planes]
        return cls(sight=tmp.sight, fps=fps, geom=tmp.geom, comps=tmp.comps,
            cuts=tmp.cuts, threads=threads, det_weights=det_weights, interpol=intepol)
    def _get_proj(self):
        if self.geom is None:
            raise ValueError("Can't project without a geometry!")
        # Backwards compatibility for old so3g
        interpol_kw = _get_interpol_args(self.interpol)
        if self.tiled:
            return MultiProj.for_tiled(
                self.geom.shape, self.geom.wcs, self.geom.tile_shape,
                active_tiles=self.active_tiles, **interpol_kw)
        else:
            return MultiProj.for_geom(self.geom.shape, self.geom.wcs, **interpol_kw)
    def _get_asm(self):
        return [so3g.proj.Assembly.attach(self.sight, fp) for fp in self.fps]

# Need a way to get PmatMultibeam into SignalMap. Either SignalMap must
# construct it itself, or it must be passed to its add_obs. We don't have
# direct access to its add_obs here though, and it's not really the
# mapmaking class' business to deal with pointing details like this.
# Cleanest solution is probably to modify the way SignalMap.add_obs
# allocates its pointing matrix, making it make a normal P unless obs
# is populated with multibeam info (though, since multibeam is a superset
# of single-beam, one could just use multibeam always. Longer term it might
# be best to just modify so3g to have multibeam support from the start.
# But for now let's keep it here.
#
# How should the multibeam info be recorded in obs? Each beam is fully
# described by a focal_plane entry. Could modify that to be a list, but
# this might break things, and it could still be good to keep the standard
# responses available. So how about making a separate multibeam structure
# that containes this more detailed description of the beam? It would then
# be populated from simpler leakage beam parameters. Let's go with that for now.

def get_pmap(obs, geom, recenter=None, **kwargs):
    if recenter:
        t0  = obs.timestamps[obs.samps.count//2]
        rot = recentering_to_quat_lonlat(*evaluate_recentering(recenter,
            ctime=t0, geom=geom, site=smutils.unarr(obs.site)))
    else: rot = None
    if "multibeam" in obs:
        return PmatMultibeam.for_tod(obs, focal_planes=obs.multibeam, geom=geom, rot=rot, **kwargs)
    else:
        return coords.pmat.P.for_tod(obs, geom=geom, rot=rot, **kwargs)



# This code has been ported over from enlib: 
# https://github.com/amaurea/enlib
# Ported over on 10/31/2023
# The original code is public domain. Modifications and additions 
# are subject to the main license in this repository.

# -*- coding: utf-8 -*-
"""This module provides a simple class for benchmarking selected parts of the
code during regular execution of a program. Usage:
    with mark("category"):
        statements
will record the time and memory usage of statement each time it is
hit in the normal course of execution of the program. For each
category, the mean and standard deviation of time and memory
are computed. The statistics are available as the module-global
stats object. str(stats) will output something like this:

      n        time          cpu           mem           leak
            mean    std   mean    std   mean    std   mean    std
 foo  200   0.03  0.020   0.03  0.021   0.76   0.00   0.00  0.000
 bar 1010   0.00  0.015   0.00  0.016   0.76   0.00   0.00  0.000

The individual entries are also avilable by indexing stats like
a dictionary: stats["foo"]["time"].mean, .std, .n, .last
return the mean, standard deviation, number of hits and
last value for the execution time for category "foo", for example.
"""
import time, numpy as np
from collections import defaultdict
from pixell import memory
try: clockfun = time.clock
except: clockfun = time.process_time

class Value:
    def __init__(self, n=0, v=0, vv=0):
        self.n  = n
        self.v  = v
        self.vv = vv
        self.last = v
    def add(self,v):
        self.n  += 1
        self.v  += v
        self.vv += v**2
        self.last = v
    @property
    def mean(self): return self.v/self.n if self.n > 0 else np.nan
    @property
    def std(self):
        if self.n == 0: return np.nan
        var = self.vv/self.n - (self.v/self.n)**2
        if var < 0: return 0.0
        else: return var**0.5
    def __repr__(self): return "Value(mean=%f, std=%f, n=%d)" % (self.mean, self.std, self.n)

class Entry(defaultdict):
    def __init__(self):
        defaultdict.__init__(self, Value)
    def __repr__(self):
        names = sorted(self)
        return "Entry("+", ".join(["%s=%s" % (name,str(self[name])) for name in names]) + ")"

class Register(defaultdict):
    def __init__(self, fmt=[("time","%6.2f","%6.3f",1),("cpu","%6.2f","%6.3f",1),("mem","%6.2f","%6.2f",2.0**30),("leak","%6.2f","%6.3f",2.0**30)]):
        defaultdict.__init__(self, Entry)
        self.info = fmt
    def add(self, name, *args):
        entry = self[name]
        for info,v in zip(self.info,args):
            entry[info[0]].add(v)
    def get(self, name):
        return self[name]["time"].mean
    def __repr__(self):
        # Sort by first ordering criterion
        if len(self) == 0: return ""
        nhit = np.array([self[name][self.info[0][0]].n    for name in self])
        vals = np.array([self[name][self.info[0][0]].mean for name in self])
        inds = np.argsort(nhit*vals)[::-1]
        names = list(self)
        names = [names[i] for i in inds]
        nhit  = nhit[inds]
        # For pretty output, determine ranges of each column
        name_dig = max([len(name) for name in names])
        nhit_dig = max([len("%d"%i) for i in nhit])
        lines = []
        # Header
        pre = " "*name_dig
        line1,line2 = pre + " %s" % "n".center(nhit_dig), pre + " "*(1+nhit_dig)
        for info in self.info:
            lens = [len(i%0) for i in info[1:3]]
            line1 += " " + info[0].center(sum(lens)+1)
            line2 += " " + " ".join("%*s" % (l,v) for l,v in zip(lens,["mean","std"]))
        lines.append(line1)
        lines.append(line2)
        for name,nh in zip(names,nhit):
            entry = self[name]
            line = "%-*s %*d" % (name_dig,name,nhit_dig,nh)
            for info in self.info:
                val = entry[info[0]]
                unit= float(info[3] if len(info) > 3 else 1)
                line += (" " + info[1] + " " + info[2]) % (val.mean/unit, val.std/unit)
            lines.append(line)
        return "\n".join(lines)
    def write(self, fname):
        with open(fname,"w") as f:
            f.write(str(self)+"\n")

stats = Register()

class mark:
    def __init__(self, name):
        self.name = name
    def __enter__(self):
        self.time1  = time.time()
        self.clock1 = clockfun()
        self.mem1   = memory.current()
    def __exit__(self, type, value, traceback):
        self.time2  = time.time()
        self.clock2 = clockfun()
        self.mem2   = memory.current()
        stats.add(self.name, self.time2 -self.time1, self.clock2-self.clock1, self.mem1, self.mem2-self.mem1)

class show:
    def __init__(self, name, display=True):
        self.name = name
        self.display = display
    def __enter__(self):
        self.time1  = time.time()
        self.clock1 = clockfun()
        self.mem1   = memory.current()
    def __exit__(self, type, value, traceback):
        self.time2  = time.time()
        self.clock2 = clockfun()
        self.mem2   = memory.current()
        self.memmax = memory.max()
        if self.display:
            print("%5.2f %5.2f %5.2f %s" % (self.time2-self.time1, self.mem2*1e-9, self.memmax*1e-9, self.name))

class dummy:
    def __init__(self, name): pass
    def __enter__(self): pass
    def __exit__(self, type, value, traceback): pass



import re
import datetime
import logging

import numpy as np

from scipy.optimize import fmin

from astropy import units
from astropy.utils import data as au_data
from skyfield import jpllib, api as skyfield_api

import so3g
from pixell import enmap

from .. import core, tod_ops, coords

logger = logging.getLogger(__name__)


# Default source list
SOURCE_LIST = ['mercury',
               'venus',
               'moon',
               'mars',
               'jupiter',
               'saturn',
               'uranus',
               'neptune',
               ('tauA', 83.6272579, 22.02159891),
               ('rcw38', 134.7805107, -47.50911231),
               ('iras16183-4958', 245.5154, -50.09168292),
               ('iras19078+0901', 287.5575891, 9.107188994),
               ('rcw122', 260.0339538, -38.95673421),
               ('cenA', 201.3625336, -43.00797508),
               ('3c279', 194.0409868, -5.79174024),
               ('3c273', 187.2775626, 2.053532671),
               ('G025.36-00.14', 279.5264042, -6.793169326),
               ('QSO_J2253+1608', 343.4952422, 16.14301323),
               ('galactic_center', -93.5833, -29.0078)]

def get_source_list_fromstr(target_source):
    """Get a source_list from SOURCE_LIST by name, or raise ValueError if not found.
    """
    for isource in SOURCE_LIST:
        if isinstance(isource, str):
            isource_name = isource
        elif isinstance(isource, tuple):
            isource_name = isource[0]
        if isource_name.lower() == target_source.lower():
            return isource
    raise ValueError(f'Source "{target_source}" not found in {SOURCE_LIST}.')

class SlowSource:
    """Class to track the time-dependent position of a slow-moving source,
    such as a Solar System planet in equatorial coordinates.
    "Slow-moving" is in relation to the time range of interest.  A
    linear approximation is good enough for Solar System, arcsecond
    accuracy, on time scales of a few hours.

    Args:
      timestamp (float): reference time, as a unix timestamp.
      ra (float): Right Ascension at reference time, in radians.
      dec (float): Declination at reference time, in radians.
      v_ra (float): rate of change of RA, in radians.
      v_dec (float): rate of change of dec, in radians.
      precision: not implemented, don't worry about it.

    """

    def __init__(self, timestamp, ra, dec, v_ra=0., v_dec=0.,
                 precision=.0001 * coords.DEG):
        self.timestamp = timestamp
        self.ra = ra
        self.dec = dec
        self.v_ra = v_ra
        self.v_dec = v_dec

    @classmethod
    def for_named_source(cls, name, timestamp):
        """Returns a SlowSource for planet ``name``, with position and
        peculiar velocity measured at time timestamp (float, unix).

        """
        dt = 3600
        ra0, dec0, distance = get_source_pos(name, timestamp)
        ra1, dec1, distance = get_source_pos(name, timestamp + dt)
        return cls(timestamp, ra0, dec0, ((ra1-ra0+np.pi) % (2*np.pi) - np.pi)/dt, (dec1-dec0)/dt)

    def pos(self, timestamps):
        """Get the (approximate) source position at the times given by the
        array of unix timestamps.

        Returns two arrays (RA, dec), the same shape as timestamps,
        both in radians.  The RA are not corrected into any particular
        branch (the mapping from timestamp to RA is continuous).

        """
        if not isinstance(timestamps, np.ndarray):
            ra, dec = self.pos(np.array([timestamps]))
            return ra[0], dec[0]
        dt = timestamps - self.timestamp
        return self.ra + self.v_ra * dt, self.dec + self.v_dec * dt


def get_scan_q(tod, planet, boresight_offset=None, refq=None):
    """Identify the point (in time and azimuth space) at which the
    specified planet crosses the boresight elevation for the
    observation in tod.  The rotation taking boresight coordinates to
    celestial coordinates at that moment in time (and pointing) is
    computed, and its conjugate is returned.

    Boresight_offset is for taking into account off-centered wafer 
    that have position offset in the focal plane, which is either None
    or a (xi, eta) tuple in radian.

    The returned rotation is useful because it represents a fixed way
    to rotate the celestial such that the target source ends up at
    (xi, eta) = 0, with the telescope scan direction parallel to xi
    and elevation parallel to eta; this is a useful system for
    measuring beam and pointing parameters.

    """
    if boresight_offset is None:
        boresight_offset = (0, 0)
    q_xieta = so3g.proj.quat.rotation_xieta(boresight_offset[0], boresight_offset[1])
    if refq is None:
        refq = so3g.proj.quat.rotation_xieta(0, 0)
    # Get reference elevation...
    el = np.median(tod.boresight.el[::10])
    az = np.median(tod.boresight.az[::10])
    t = (tod.timestamps[0] + tod.timestamps[-1]) / 2
    if isinstance(planet, (list, tuple)):
        _, ra, dec = planet
        planet = SlowSource(t, float(ra) * coords.DEG,
                                        float(dec) * coords.DEG)
    else:
        planet = SlowSource.for_named_source(planet, t)

    def scan_q_model(t, az, el, planet):
        csl = so3g.proj.CelestialSightLine.az_el(
            t, az, el, weather='typical', site='so')
        ra0, dec0 = planet.pos(t)
        return csl.Q, ~so3g.proj.quat.rotation_lonlat(ra0, dec0) * csl.Q * q_xieta

    def distance(p):
        dt, daz = p
        q, qnet = scan_q_model(t + dt, az + daz, el, planet)
        lon, lat, phi = so3g.proj.quat.decompose_lonlat(qnet)
        return 90 * coords.DEG - lat

    p0 = np.array([0, 0])
    X = fmin(distance, p0, disp=0, full_output=1)
    p, fopt, n_iter, n_calls, warnflag = X
    if warnflag != 0:
        logger.warning('Source-scan solver failed to converge or otherwise '
                       f'complained!  warnflag={warnflag}')
    q, qnet = scan_q_model(t+p[0], az+p[1], el, planet)
    psi = so3g.proj.quat.decompose_xieta(qnet)[2][0]
    ra, dec = planet.pos(t+p[0])
    rot = ~so3g.proj.quat.rotation_lonlat(ra, dec, psi)
    return {'rot': rot,
            'timestamp': t+p[0],
            'az': az+p[1],
            'el': el,
            'ra': ra,
            'dec': dec,
            'psi': psi,
            'planet': planet}


def get_scan_P(tod, planet, boresight_offset=None, refq=None, res=None, size=None, **kw):
    """Get a standard Projection Matrix targeting a planet (or some
    interesting fixed position), in source-scan coordinates.

    Returns a Projection Matrix and the output from get_scan_q.

    """
    logger.debug(f'get_scan_P: init for {planet}')

    if res is None:
        res = 0.01 * coords.DEG
    X = get_scan_q(tod, planet, boresight_offset=boresight_offset, refq=refq)
    rot = so3g.proj.quat.rotation_lonlat(0, 0) * X['rot']
    wcs_kernel = coords.get_wcs_kernel('tan', 0., 0., res=res)

    logger.debug(f'get_scan_P: getting projection matrix for {wcs_kernel}.')
    P = coords.P.for_tod(tod, rot=rot, wcs_kernel=wcs_kernel, **kw)
    if size is not None:
        # Trim to a local square
        mz = P.zeros(comps='T').submap([[-size/2, size/2], [size/2, -size/2]])
        P.geom = enmap.Geometry(shape=mz.shape, wcs=mz.wcs)
    return P, X


def get_horizon_P(tod, az, el, receiver_fixed=False, **kw):
    """Get a standard Projection Matrix targeting arbitrary source (for example
    drone) in horizon coordinates.

    Args:
      tod: AxisManager of the observation
      az: azimuth of the target source (rad)
      el: elevation of the target source (rad)
      receiver_fixed:
          If true, rotate the source centered coordinate to be receiver fixed.
          This should be True for receiver fixed beam measurement.
          This should be False for pol angle calibration with source on drone.

    Return:
      a Projection Matrix

    """
    sight = so3g.proj.CelestialSightLine.for_horizon(
        tod.timestamps,
        tod.boresight.az,
        tod.boresight.el,
        tod.boresight.roll
    )
    pq = so3g.proj.quat.rotation_lonlat(-az, el)
    if receiver_fixed:
        rot = so3g.proj.quat.euler(2, -tod.boresight.roll)
        sight.Q = so3g.proj.quat.rotation_lonlat(0, 0) * rot * ~pq * sight.Q
    else:
        sight.Q = so3g.proj.quat.rotation_lonlat(0, 0) * ~pq * sight.Q
    P = coords.P.for_tod(tod, sight, **kw)
    return P


def filter_for_sources(tod=None, signal=None, source_flags=None,
                       n_modes=10, low_pass=None,
                       wrap=None, edge_guard=None):
    """Mask and gap-fill the signal at samples flagged by source_flags.
    Then PCA the resulting time ordered data.  Restore the flagged
    signal, remove the strongest modes from PCA.

    If signal is not passed in tod.signal will be modified directly.
    To leave tod.signal intact, pass in signal=tod.signal.copy().

    Args:
      tod: AxisManager from which defaults will be drawn.
      signal: Time-ordered data to operate on.  Defaults to
        tod.signal.
      source_flags: RangesMatrix to use for source flagging.  Defaults
        to tod.source_flags.
      low_pass: Frequency, in Hz, at which to apply low pass filter to
        signal.  If None, no filtering is done.  You can pass in a
        filter from tod_ops.filters if you want.
      n_modes (int): Number of eigenmodes to remove... interface
        subject to change.
      wrap (str): If specified, the result will be stored at
        tod[wrap].
      edge_guard (int): Number of samples at the beginning and end of the flags to change them False.
        Default is None. (Nothing happens.)

    Returns:
      The filtered signal.

    """
    if source_flags is None and tod is not None:
        source_flags = tod.source_flags
    elif source_flags is None and tod is None:
        raise ValueError("Must provide source_flags if tod is None")
    if signal is None and tod is not None:
        signal = tod.signal
    elif signal is None and tod is None:
        raise ValueError("Must provide signal if tod is None")
    if signal is None:
        raise ValueError("signal somehow still None!")
    
    if edge_guard is not None:
        if isinstance(edge_guard, int):
            edgebl = np.zeros(source_flags.shape[-1], dtype=bool)
            edgebl[edge_guard:-edge_guard] = True
            edgerm = so3g.proj.ranges.RangesMatrix.from_mask(edgebl)
            for iflag in source_flags:
                iflag.intersect(edgerm)
        else:
            raise ValueError("edge_guard must be an int.")

    # Get a reasonable gap fill.
    signal_pca = signal.copy()
    gaps = tod_ops.get_gap_fill(tod, signal=signal_pca, flags=source_flags)
    gaps.swap(tod, signal=signal_pca)

    # Low pass filter?
    if low_pass is not None:
        if tod is None:
            raise ValueError("tod cannot be None when lowpassing")
        if isinstance(low_pass, tod_ops.filters._chainable):
            filt = low_pass
        else:
            filt = tod_ops.filters.low_pass_butter4(low_pass)

        n_det, n = signal.shape
        rfft = tod_ops.fft_ops.RFFTObj.for_shape(n_det, n, 'BOTH')
        rfft.a[:] = signal_pca
        rfft.t_forward()
        times = tod.timestamps
        delta_t = (times[-1]-times[0])/(tod.samps.count - 1)
        freqs = np.fft.rfftfreq(n, delta_t)
        filt.apply(freqs, tod, target=rfft.b)
        signal_pca = rfft.t_backward()
        del rfft

    # Measure TOD means (after gap fill, low pass, etc).
    if isinstance(n_modes, str) and n_modes == 'all':
        # Don't overthink that.
        signal -= signal_pca

    else:
        # Make sure PCA decomposition targets (signal_pca) are mean 0.
        # It's convenient to remove the same levels from signal now,
        # too.
        levels = signal_pca.mean(axis=1)
        signal_pca -= levels[:, None]
        signal -= levels[:, None]

        # Get PCA model and discard the source vectors.
        pca = tod_ops.pca.get_pca_model(
            tod, signal=signal_pca, n_modes=n_modes)
        del signal_pca

        # Remove the PCA model.
        tod_ops.pca.add_model(tod, pca, -1, signal=signal)

    if wrap and tod is not None:
        tod.wrap(wrap, signal, [(0, 'dets'), (1, 'samps')])
    return signal

def _get_astrometric(source_name, timestamp, site="_default", planets=None):
    """
    Derive skyfield's Astrometric object of a celestial source at a
    specific timestamp and observing site, which is used to derive
    radec/azel in get_source_pos/get_source_azel.

    Args:
      source_name: Planet name; in capitalized format, e.g. "Jupiter",
        or fixed source specification.
      timestamp: unix timestamp.
      site (str or so3g.proj.EarthlySite): if this is a string, the
        site will be looked up in so3g.proj.SITES dict.
      planets (SpiceKernel): the ephemeris object for
        skyfield.jpllib). If not passed in, a sensible default
        ephemeris is used.

    Returns:
      planets: SpiceKernel object (for re-use)
      astrometric: skyfield's astrometric object

    Notes:
      The SpiceKernel holds the ephemeris file open. To close that
      file, run planets.close().  Deleting the object is not enough,
      as the garbage collection can be lazy.  After .close(), the
      kernel can't be used for ephemeris stuff. Some operations on the
      returned "astrometric" object will try to use the kernel, and
      will fail. So keep planets around until you're done computing on
      astrometric, then call close().

      If you're doing lots of astrometric stuff, you can get the
      kernel once and re-use it.

      If the default ephemeris file (de421.bsp) is not found, it will
      be downloaded (16M) and cached.

    """
    if planets is None:
        # Get the ephemeris
        de_filename = core.get_local_file("de421.bsp")
        planets = jpllib.SpiceKernel(de_filename)
    else:
        assert len(planets.codes)  # Someone .closed this eph!

    for k in [
        source_name,
        source_name + " barycenter",
    ]:
        try:
            target = planets[k]
            break
        except (ValueError, KeyError):
            pass
    else:
        options = list(planets.names().values())
        raise ValueError(
            f'Failed to find a match for "{source_name}" in ephemeris: {options}'
        )

    if isinstance(site, str):
        site = so3g.proj.SITES[site]

    observatory = site.skyfield_site(planets)

    timescale = skyfield_api.load.timescale()
    sf_timestamp = timescale.from_datetime(
        datetime.datetime.fromtimestamp(timestamp, tz=skyfield_api.utc)
    )
    astrometric = observatory.at(sf_timestamp).observe(target)
    return planets, astrometric


def get_source_pos(source_name, timestamp, site='_default'):
    """Get the equatorial coordinates of a planet (or fixed-position
    source, see note) at some time.  Returns the apparent position,
    accounting for geographical position on earth, but assuming no
    atmospheric refraction.
    
    Args:
      source_name: Planet name; in capitalized format, e.g. "Jupiter",
        or fixed source specification.
      timestamp: unix timestamp.
      site (str or so3g.proj.EarthlySite): if this is a string, the
        site will be looked up in so3g.proj.SITES dict.

    Returns:
      ra (float): in radians.
      dec (float): in radians.
      distance (float): in AU.

    Note:

      Before checking in the ephemeris, the source_name will be
      matched against a regular expression and if it has the format
      'Jxxx[+-pmn]yyy', where xxx and yyy are decimal numbers, then a
      fixed-position source at RA,Dec = xxx,yyy in degrees will be
      processed.  In that case, the distance is returned as Inf.

    """
    # Check against fixed-position template...
    m = re.match(
        r'[jJ](?P<ra_deg>\d+(\.\d*)?)(?P<dec_sign>[+-pmn])(?P<dec_deg>\d+(\.\d*)?)', source_name)
    if m:
        sign = (-1)**(m['dec_sign'] in ['-', 'm', 'n'])
        ra, dec = float(m['ra_deg']) * \
            coords.DEG, sign * float(m['dec_deg']) * coords.DEG
        return ra, dec, float('inf')
    
    # Derive from skyfield astrometric object
    planets, amet0 = _get_astrometric(source_name, timestamp, site)
    ra, dec, distance = amet0.radec()
    planets.close()
    return ra.to(units.rad).value, dec.to(units.rad).value, distance.to(units.au).value


def get_source_azel(source_name, timestamp, site='_default'):
    """
    Get the apparent azimuth and elevation of a celestial source at a 
    specific timestamp and observing site. Returns the apparent position,
    accounting for geographical position on earth, but assuming no
    atmospheric refraction.

    Args:
        source_name: Planet name; in capitalized format, e.g. "Jupiter"
        timestamp (float): The Unix timestamp representing the time for 
          which to calculate azimuth and elevation.
        site (str or so3g.proj.EarthlySite): if this is a string, the
        site will be looked up in so3g.proj.SITES dict.

    Returns:
      az (float): in radians.
      el (float): in radians.
      distance (float): in AU.
    """
    planets, amet0 = _get_astrometric(source_name, timestamp, site)
    el, az, distance = amet0.apparent().altaz()
    planets.close()
    return az.to(units.rad).value, el.to(units.rad).value, distance.to(units.au).value


def get_nearby_sources(tod=None, source_list=None, distance=1.):
    """Identify solar system objects (especially "planets") that might be
    within a TOD's scan footprint.

    Arguments:
      tod (AxisManager): The data to check.  Needs to have
        focal_plane, boresight, timestamps.
      source_list (list or None): A list of source names or None to
        use a default list.  Use simple planet names in lower case
        (e.g. ['uranus']), or tuples with source name, RA, and dec in
        degrees (e.g. [('tau_a', 83.63, 22.01)]).
      distance (float): Maximum distance from the source center, in
        degrees, to consider as "within the footprint".  (This should
        be at least the sum of the beam radius and the planet radius
        ... though there's usually no harm in going a bit larger than
        that.)

    Returns:
      List of tuples (source_name, SlowSource) that satisfy the
      "nearby" condition.

    """
    # Make a full sky map with not very many pixels.
    shape, wcs = enmap.fullsky_geometry(res=2 * coords.DEG, proj='car')

    # Sight line
    sight = so3g.proj.CelestialSightLine.az_el(
        tod.timestamps, tod.boresight.az, tod.boresight.el,
        site='so', weather='typical')

    # One central detector
    xieta0, R, _ = coords.helpers.get_focal_plane_cover(tod, 0)
    fp = so3g.proj.FocalPlane.from_xieta([xieta0[0]], [xieta0[1]])

    asm = so3g.proj.Assembly.attach(sight, fp)
    p = so3g.proj.Projectionist.for_geom(shape, wcs)
    w = p.to_map(np.zeros((1, len(tod.timestamps)),
                 'float32')+1, asm, comps='T')
    w = enmap.enmap(w, wcs=wcs)

    if source_list is None:
        source_list = SOURCE_LIST

    positions = []
    for source_name in source_list:
        t = tod.timestamps[0]
        if isinstance(source_name, (list, tuple)):
            source_name, ra, dec = source_name
            sl = coords.planets.SlowSource(t, float(ra) * coords.DEG,
                                           float(dec) * coords.DEG)
        else:
            sl = coords.planets.SlowSource.for_named_source(source_name, t)
        x = w.distance_from([[sl.dec], [sl.ra]])
        md = x[w[0] != 0].min()
        logger.debug(('Source {:12} is at ({:8.4f},{:8.4f}); '
                      'that is {:5.2f} degrees off footprint.').format(
                          source_name, sl.ra / coords.DEG,
                          sl.dec / coords.DEG, md/coords.DEG))
        if md < (R * 1.1 + distance * coords.DEG):
            positions.append((source_name, sl))
    return positions


def compute_source_flags(tod=None, P=None, mask=None, wrap=None,
                         center_on=None, res=None, max_pix=4e6):
    """Process masking instructions and create RangesMatrix that flags
    samples in the TOD that are within the masked region.  This
    masking makes use of a map with the footprint encoded in P, so
    flagging boundaries will correspond to pixel edges.

    The interface for "mask" is subject to change -- use of a simple
    text file is interim.

    Args:
      tod (AxisManager): the observation.
      P (Projection Matrix): if passed in, must include a map geom
        (shape and WCS).  If None, will be created from get_scan_P
        using center_on and res parameters.
      mask: source masking instructions (see note).
      wrap: key in tod at which to store the result.
      res: If P is None, sets the target mask map resolution
        (radians).
      max_pix: If P is None, this sets the maximum acceptable number
        of pixels for the mask map.  This is to catch cases where an
        incorrect source has been passed in, for example, leading to a
        weird map footprint

    Returns:
      RangesMatrix marking the samples inside the masked region.

    Notes:
      The mask can be a dict or a list of dicts.  Each dict must be of
      the form::

        {'shape': 'circle',
         'xyr': (XI, ETA, R)}

      where R is the radius of the circular mask, and XI and ETA are
      the center of the circle, all in degrees.

    """
    if P is None:
        logger.info('Getting Projection Matrix ...')
        P, X = get_scan_P(tod, center_on, res=res, comps='T')
        shape, wcs = tuple(P.geom)
        if shape[0] * shape[1] > max_pix:
            raise ValueError(f'Mask map too large: {shape}')

    if isinstance(mask, str):
        # Assume it's a filename, and file is simple columns of (x, y,
        # radius) in deg.  (Deprecated!)
        mask = [{'xyr': list(map(float, line.split()))}
                for line in open(mask)]

    mask_map = P.zeros()
    _add_to_mask(mask, mask_map)
    a = P.from_map(mask_map)
    source_flags = so3g.proj.RangesMatrix(
        [so3g.proj.Ranges.from_mask(r != 0) for r in a])

    if wrap:
        assert tod is not None, "Pass in a tod to 'wrap' the output."
        tod.flags.wrap(wrap, source_flags, [(0, 'dets'), (1, 'samps')])
    return source_flags


def _add_to_mask(req, mask_map):
    # Helper for compute_source_flags.
    if req is None:
        raise ValueError(f'Requested mask is None.  For no mask, pass [].')
    if isinstance(req, (list, tuple)):
        for _r in req:
            # Also, maybe test this somehow?
            _add_to_mask(_r, mask_map)
    elif isinstance(req, dict):
        shape = req.get('shape', 'circle')
        if shape == 'circle':
            x, y, r = req['xyr']
            d = enmap.distance_from(mask_map.shape, mask_map.wcs,
                                    [[y * coords.DEG], [x * coords.DEG]])
            mask_map += 1. * (d < r * coords.DEG)
        else:
            raise ValueError(f'Unknown shape="{shape}" in mask request.')
    else:
        raise ValueError(f'Weird mask request: {req}')


def load_detector_splits(tod=None, filename=None, dataset=None,
                         source=None, wrap=None):
    """Convert a partition of detectors into a dict of disjoint
    RangesMatrix objects; such an object can be passed to make_map()
    to efficiently make detector-split maps.

    The "detector split" data can be read from an HDF5 dataset, or
    passed in directly as an AxisManager.

    Args:
      tod (AxisManager): This is required, to get the list of dets and
        the samps count.
      filename (str): The HDF filename, or filename:dataset.
      dataset (str): The HDF dataset (if not passed in with filename).
      source (array, ResultSet or AxisManager): If not None, then
        filename and dataset are ignored and this object is processed
        (as though it had just been loaded from HDF).
      wrap (str): If not None, the address in tod where to store the
        loaded split data.

    The format of detector splits, in an HDF5 dataset, is as one would
    write from a ResultSet with columns ['dets:name', 'group'] (both
    str).  All detectors sharing a value in the group column will
    grouped together and the group label will be that value.

    If passing in "source" directly as a ResultSet, it should have
    columns 'dets:name' and 'group'; if as an AxisManager then it
    should have a 'dets' axis and a vector 'group' with shape
    ('dets',) providing the group name for each detector.  If it is a
    numpy array, it is assumed to correspond one-to-one with the .dets
    axis of TOD and the array gives the group name for each detector.

    Returns:
      data_splits (dict of RangesMatrix): Each entry of the dict is a
        RangesMatrix that can be interpreted as cuts to apply during
        mapmaking.  In this case the RangesMatrix will simply mark
        each detector as either fully cut (flagged) or fully uncut.

    """
    from sotodlib.io import metadata

    if source is None:
        if dataset is None:
            filename, dataset = filename.split(':')
        source = metadata.read_dataset(filename, dataset)
    if isinstance(source, np.ndarray):
        source, _s = core.AxisManager(tod.dets), source
        source.wrap('group', _s)
    elif isinstance(source, metadata.ResultSet):
        di = core.metadata.loader.unconvert_det_info(tod.det_info)
        source = core.metadata.loader.broadcast_resultset(
            source, di, axis_key='name')
    else:
        source = source.copy()  # is this an AxisManager?
    source.restrict_axes([tod.dets])
    if wrap:
        tod.wrap(wrap, source)
    yes = so3g.proj.RangesMatrix.zeros(tod.signal.shape[1])
    flags = {}
    for group in source['group']:
        if group in flags:
            continue
        flags[group] = so3g.proj.RangesMatrix.ones((tod.signal.shape))
        for i in (source['group'] == group).nonzero()[0]:
            flags[group].ranges[i] = yes
    return flags


def get_det_weights(tod, signal=None, wrap=None,
                    outlier_clip=None):
    """Compute detector weights, based on variance of signal.  If
    outlier_clip is set, it is used to trim outliers.  See code for
    details, but think of it as the number of stdev away from the mean
    weight that will be kept, and try 2.5.

    Returns:
      A det_weights array compatible with projection matrix functions.

    """
    if signal is None:
        signal = tod.signal
    det_weights = np.zeros(signal.shape[:-1])
    sigmas = signal.std(axis=-1)
    ok = sigmas != 0
    det_weights[ok] = sigmas[ok]**-2
    if outlier_clip:
        qs = np.quantile(det_weights[ok], [.16, .84])  # "1 sigma" lims
        q0, dq = qs.mean(), np.diff(qs)[0]/2
        ok *= ((q0 - outlier_clip * dq < det_weights) *
               (det_weights < q0 + outlier_clip * dq))
    return det_weights * ok


def write_det_weights(tod, filename, dataset, det_weights=None):
    """Save detector weights to an HDF file dataset.

    Args:
      tod (AxisManager): provides the dets axis, and det_weights if
        not passed explicitly.
      filename (str or h5py.File): destination filename (or open File)
      dataset (str): address in the HDF5 file to put the result (it
        will be overwritten if exists already).
      det_weights (array): the weights to write; must be in
        correspondence with tod.dets.  Defaults to tod.det_weights.

    Returns:
      ResultSet with the info that was written.

    """
    from sotodlib.io import metadata
    if det_weights is None:
        det_weights = tod.det_weights
    rs = core.metadata.ResultSet(['dets:name', 'det_weights'])
    rs.rows.extend(list(zip(tod.dets.vals, det_weights)))
    metadata.write_dataset(rs, filename, dataset, overwrite=True)
    return rs


def make_map(tod, center_on=None, scan_coords=True, thread_algo=False,
             res=0.01*coords.DEG, size=None, wcs_kernel=None, comps='TQU',
             signal=None,
             det_weights=None,
             filename=None, source_flags=None, cuts=None,
             data_splits=None,
             low_pass=None, n_modes=10,
             eigentol=1e-3, info={}):
    """Make a compact source map from the TOD.  Specify filename to write
    things to disk; this should be a format string, for example
    '{obs_id}_{map}.fits', where 'map' will be given values of
    ['binned', 'solved', 'weights'] and any other keys (obs_id in this
    case) must be passed through info.

    """
    assert (center_on is not None)  # Pass in the source name, e.g. 'uranus'

    # Test the map format string...
    if filename is not None:
        try:
            filename.format(map='binned', **info)
        except:
            raise ValueError('Failed to process filename format "%s" with info=%s' %
                             (filename, info))

    class MmTimer(coords.Timer):
        PRINT_FUNC = logger.info
        FMT = 'make_map: {msg}: {elapsed:.3f} seconds'

    if signal is None:
        signal = tod.signal

    if thread_algo == 'none':
        thread_algo = False

    with MmTimer('setup thread_algo=%s' % thread_algo):
        if scan_coords:
            P, X = coords.planets.get_scan_P(tod, center_on, res=res, size=size,
                                             comps=comps, cuts=cuts,
                                             threads=thread_algo)
        else:
            planet = coords.planets.SlowSource.for_named_source(
                center_on, tod.timestamps[0])
            ra0, dec0 = planet.pos(tod.timestamps.mean())
            wcsk = coords.get_wcs_kernel('tan', ra0, dec0, res=res)
            P = coords.P.for_tod(tod, comps=comps,
                                 cuts=cuts,
                                 threads=thread_algo,
                                 wcs_kernel=wcsk)
    with MmTimer('get_proj_threads'):
        P._get_proj_threads()

    with MmTimer('filter for sources'):
        filter_for_sources(tod, signal=signal, source_flags=source_flags,
                           low_pass=low_pass, n_modes=n_modes)

    if det_weights is None:
        det_weights = get_det_weights(tod, signal=signal, outlier_clip=2.)

    if data_splits is not None:
        # Clear P's internal cuts as we'll be modifying those to pass
        # in directly.
        base_cuts, P.cuts = P.cuts, None
        output = {
            'P': P,
            'det_weights': det_weights,
            'splits': {}
        }

        # Write out _map and _weights for each group.
        for group_label, group_cuts in data_splits.items():
            logger.info(f'Mapping split "{group_label}"')
            if base_cuts is not None:
                group_cuts = group_cuts + base_cuts
            with MmTimer('getting weights'):
                w = P.to_weights(cuts=group_cuts, det_weights=det_weights)
            with MmTimer('getting map and applying inverse weights'):
                m = P.remove_weights(
                    tod=tod, signal=signal, weights_map=w, cuts=group_cuts,
                    det_weights=det_weights, eigentol=eigentol)
            output['splits'][group_label] = {
                'binned': None,
                'weights': w.astype('float32'),
                'solved': m.astype('float32'),
            }
            if filename is not None:
                m.astype('float32').write(
                    filename.format(map=f'{group_label}_map', **info))
                w.astype('float32').write(
                    filename.format(map=f'{group_label}_weights', **info))
        return output

    with MmTimer('project signal and weight maps'):
        map1 = P.to_map(tod, signal=signal, det_weights=det_weights)
        wmap1 = P.to_weights(det_weights=det_weights)

    with MmTimer('compute and apply inverse weights map'):
        iwmap1 = P.to_inverse_weights(weights_map=wmap1, eigentol=eigentol)
        map1b = P.remove_weights(map1, inverse_weights_map=iwmap1)

    if filename is not None:
        with MmTimer('Write out'):
            map1b.write(filename.format(map='solved', **info))
            if filename.format(map='x', **info) != filename.format(map='y', **info):
                map1.write(filename.format(map='binned', **info))
                wmap1.write(filename.format(map='weights', **info))
                write_det_weights(tod, filename.format(map='detweights', **info).
                                  replace('.fits', '.h5'), 'detweights',
                                  det_weights=det_weights)

    return {'binned': map1,
            'weights': wmap1,
            'solved': map1b,
            'P': P,
            'det_weights': det_weights,
            }



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Hardware visualization tools.
"""

import astropy.units as u
import numpy as np
import quaternionarray as qa
import warnings

import toast

from so3g.proj import quat

from .core.hardware import LAT_COROTATOR_OFFSET

from .sim_hardware import sim_nominal


default_band_colors = {
    "LAT_f030": (0.4, 0.4, 1.0, 0.2),
    "LAT_f040": (0.4, 0.4, 1.0, 0.2),
    "LAT_f090": (0.4, 1.0, 0.4, 0.2),
    "LAT_f150": (0.4, 1.0, 0.4, 0.2),
    "LAT_f230": (1.0, 0.4, 0.4, 0.2),
    "LAT_f290": (1.0, 0.4, 0.4, 0.2),
    "SAT_f030": (0.4, 0.4, 1.0, 0.2),
    "SAT_f040": (0.4, 0.4, 1.0, 0.2),
    "SAT_f090": (0.4, 1.0, 0.4, 0.2),
    "SAT_f150": (0.4, 1.0, 0.4, 0.2),
    "SAT_f230": (1.0, 0.4, 0.4, 0.2),
    "SAT_f290": (1.0, 0.4, 0.4, 0.2),
}


def set_matplotlib_pdf_backend():
    """Set the matplotlib backend to PDF

    This is necessary to render properly the focal plane plots"""

    import matplotlib

    matplotlib.use("pdf")
    import matplotlib.pyplot as plt

    return plt


def plot_detectors(
    dets,
    outfile,
    width=None,
    height=None,
    labels=False,
    bandcolor=None,
    xieta=False,
    lat_corotate=True,
    lat_elevation=None,
    show_centers=False,
):
    """Visualize a dictionary of detectors.

    This makes a simple plot of the detector positions on the projected
    focalplane.  The size of detector circles are controlled by the detector
    "fwhm" key, which is in arcminutes.  If the bandcolor is specified it will
    override the defaults.

    Args:
        outfile (str): Output PDF path.
        dets (dict): Dictionary of detector properties.
        width (float): Width of plot in degrees (None = autoscale).
        height (float): Height of plot in degrees (None = autoscale).
        labels (bool): If True, label each detector.
        bandcolor (dict, optional): Dictionary of color values for each band.
        xieta (bool):  If True, plot in Xi / Eta / Gamma coordinates rather
            than focalplane X / Y / Z.
        show_centers (bool):  If True, label pixel centers.

    Returns:
        None

    """
    try:
        plt = set_matplotlib_pdf_backend()
    except:
        wmsg = (
            "Couldn't set the PDF matplotlib backend, "
            "focal plane plots will not render properly, "
            "proceeding with the default matplotlib backend"
        )
        warnings.warn(wmsg)
        import matplotlib.pyplot as plt

    xaxis, yaxis, zaxis = np.eye(3)

    # Get wafer to telescope map
    hw = sim_nominal()
    wfmap = hw.wafer_map()

    n_det = len(dets)
    detnames = list(dets.keys())
    quats = np.array(
        [dets[detnames[x]]["quat"] for x in range(n_det)], dtype=np.float64
    )

    # Skip bolometers that do not have a proper quaternion
    good = np.isfinite(quats[:, 0])
    bad = np.logical_not(good)
    nbad = np.sum(bad)
    if nbad != 0:
        print(f"Skipping {nbad} detectors without quaternions")
        for detname, is_bad in zip(detnames, bad):
            if is_bad:
                del dets[detname]
        detnames = list(dets.keys())
        quats = quats[good]
        n_det = len(dets)

    lat_rot = None
    lat_elstr = ""

    have_lat = False
    for d in range(n_det):
        dn = detnames[d]
        tele = wfmap["telescopes"][dets[dn]["wafer_slot"]]
        if tele == "LAT":
            have_lat = True

    if have_lat:
        if not lat_corotate:
            if lat_elevation is None:
                raise RuntimeError("Must specify elevation if not co-rotating")
            lat_elstr = f"({lat_elevation.to_value(u.degree):0.1f} Degrees Elevation)"
            lat_ang = lat_elevation.to_value(u.rad) - LAT_COROTATOR_OFFSET.to_value(
                u.rad
            )
            lat_rot = qa.rotation(zaxis, lat_ang)
        for d in range(n_det):
            dn = detnames[d]
            tele = wfmap["telescopes"][dets[dn]["wafer_slot"]]
            if tele == "LAT" and lat_rot is not None:
                quats[d] = qa.mult(lat_rot, quats[d])

    if n_det == 0:
        raise RuntimeError("No detectors specified")

    # arc_factor returns a scaling that can be used to reproject (X,
    # Y) to units corresponding to the angle subtended from (0,0,1) to
    # (X, Y, sqrt(X^2 + Y^2)).  This is called "ARC" (Zenithal
    # Equidistant) projection in FITS.
    def arc_factor(x, y):
        r = (x**2 + y**2) ** 0.5
        if r < 1e-6:
            return 1.0 + r**2 / 6
        return np.arcsin(r) / r

    if xieta:
        # Plotting as seen from observer in Xi / Eta / Gamma
        xi, eta, gamma = toast.instrument_coords.quat_to_xieta(quats)
        # xi, eta, gamma = quat.decompose_xieta(quats)

        detx = {
            detnames[k]: xi[k] * 180.0 / np.pi * arc_factor(xi[k], eta[k])
            for k in range(n_det)
        }
        dety = {
            detnames[k]: eta[k] * 180.0 / np.pi * arc_factor(xi[k], eta[k])
            for k in range(n_det)
        }

        for didx, dn in enumerate(detnames):
            wf = dets[dn]["wafer_slot"]
            px = int(dets[dn]["pixel"])
            if px == 0:
                q = quats[didx]
                dd = qa.rotate(q, zaxis)

        # In Xi / Eta coordinates, gamma is measured clockwise from line of
        # decreasing elevation.  Here we convert into visualization X/Y
        # coordinatates measured counter clockwise from the X axis.
        polangs = {detnames[k]: 1.5 * np.pi - gamma[k] for k in range(n_det)}
    else:
        # Plotting in focalplane X / Y / Z coordinates.
        # Compute direction and orientation vectors
        dir = qa.rotate(quats, zaxis)
        orient = qa.rotate(quats, xaxis)

        small = np.fabs(1.0 - dir[:, 2]) < 1.0e-12
        not_small = np.logical_not(small)
        xp = np.zeros(n_det, dtype=np.float64)
        yp = np.zeros(n_det, dtype=np.float64)

        mag = np.arccos(dir[not_small, 2])
        ang = np.arctan2(dir[not_small, 1], dir[not_small, 0])
        xp[not_small] = mag * np.cos(ang)
        yp[not_small] = mag * np.sin(ang)

        polangs = {
            detnames[k]: np.arctan2(orient[k, 1], orient[k, 0]) for k in range(n_det)
        }

        detx = {
            detnames[k]: xp[k] * 180.0 / np.pi * arc_factor(xp[k], yp[k])
            for k in range(n_det)
        }
        dety = {
            detnames[k]: yp[k] * 180.0 / np.pi * arc_factor(xp[k], yp[k])
            for k in range(n_det)
        }

    if (width is None) or (height is None):
        # We are autoscaling.  Compute the angular extent of all detectors
        # and add some buffer.
        _y = np.array(list(dety.values()))
        _x = np.array(list(detx.values()))
        wmin, wmax = _x.min(), _x.max()
        hmin, hmax = _y.min(), _y.max()
        wbuf = 0.2 * (wmax - wmin)
        hbuf = 0.2 * (hmax - hmin)
        wmin -= wbuf
        wmax += wbuf
        hmin -= hbuf
        hmax += hbuf
        width = wmax - wmin
        height = hmax - hmin
        half_width = 0.5 * width
        half_height = 0.5 * height
    else:
        half_width = 0.5 * width
        half_height = 0.5 * height
        wmin = -half_width
        wmax = half_width
        hmin = -half_height
        hmax = half_height

    wafer_centers = dict()
    if labels:
        # We are plotting labels and will want to plot a label for each
        # wafer.  To decide where to place the label, we find the average location
        # of all detectors from each wafer and put the label there.
        temp_centers = dict()
        for d, props in dets.items():
            dwslot = props["wafer_slot"]
            if dwslot not in temp_centers:
                temp_centers[dwslot] = {"x": list(), "y": list()}
            temp_centers[dwslot]["x"].append(detx[d])
            temp_centers[dwslot]["y"].append(dety[d])
        for k in list(temp_centers.keys()):
            xcenter = np.mean(temp_centers[k]["x"])
            ycenter = np.mean(temp_centers[k]["y"])
            ysize = 0.25 * np.std(np.array(temp_centers[k]["y"]) - ycenter)
            tube_slot = wfmap["tube_slots"][k]
            tube_props = hw.data["tube_slots"][tube_slot]
            tube_wafer_indx = hw.data["wafer_slots"][k]["tube_index"]
            ufm_slot = tube_props["wafer_ufm_slot"][tube_wafer_indx]
            if "wafer_ufm_loc" in tube_props:
                ufm_loc = tube_props["wafer_ufm_loc"][tube_wafer_indx]
                ltext = f"{k}:ws{ufm_slot}:{ufm_loc}"
            else:
                ltext = f"{k}:ws{ufm_slot}"
            wafer_centers[k] = ((xcenter, ycenter), ysize, ltext)

    if bandcolor is None:
        bandcolor = default_band_colors
    wfigsize = 10.0
    hfigsize = wfigsize * (height / width)
    figdpi = 75
    hfigpix = int(figdpi * hfigsize)
    hpixperdeg = hfigpix / height

    fig = plt.figure(figsize=(wfigsize, hfigsize), dpi=figdpi)
    ax = fig.add_subplot(1, 1, 1)

    ax.set_xlim([wmin, wmax])
    ax.set_ylim([hmin, hmax])
    if xieta:
        ax.set_xlabel(r"Boresight $\xi$ Degrees", fontsize="large")
        ax.set_ylabel(r"Boresight $\eta$ Degrees", fontsize="large")
    else:
        ax.set_xlabel("Boresight X Degrees", fontsize="large")
        ax.set_ylabel("Boresight Y Degrees", fontsize="large")

    # Draw wafer labels in the background
    if labels:
        # The font size depends on the wafer size ... but keep it
        # between (0.01 and 0.1) times the size of the figure.
        for k, (center, size, ltext) in wafer_centers.items():
            fontpix = np.clip(0.7 * size * hpixperdeg, 0.01 * hfigpix, 0.10 * hfigpix)
            if fontpix < 1.0:
                fontpix = 1.0
            ax.text(
                center[0],
                center[1] + fontpix / hpixperdeg,
                ltext,
                color="k",
                fontsize=fontpix,
                horizontalalignment="center",
                verticalalignment="center",
                zorder=100,
                bbox=dict(fc="white", ec="none", pad=0.2, alpha=1.0),
            )

    for d, props in dets.items():
        band = props["band"]
        pixel = props["pixel"]
        pol = props["pol"]
        quat = np.array(props["quat"]).astype(np.float64)
        fwhm = props["fwhm"]

        # radius in degrees
        detradius = 0.5 * fwhm / 60.0

        # Position and polarization angle
        xpos, ypos = detx[d], dety[d]
        polang = polangs[d]

        detface = bandcolor[band]

        circ = plt.Circle(
            (xpos, ypos),
            radius=detradius,
            fc=detface,
            ec="black",
            linewidth=0.05 * detradius,
        )
        ax.add_artist(circ)

        ascale = 1.5

        xtail = xpos - ascale * detradius * np.cos(polang)
        ytail = ypos - ascale * detradius * np.sin(polang)
        dx = ascale * 2.0 * detradius * np.cos(polang)
        dy = ascale * 2.0 * detradius * np.sin(polang)

        detcolor = "black"
        if pol == "A":
            detcolor = (1.0, 0.0, 0.0, 1.0)
        if pol == "B":
            detcolor = (0.0, 0.0, 1.0, 1.0)

        ax.arrow(
            xtail,
            ytail,
            dx,
            dy,
            width=0.1 * detradius,
            head_width=0.3 * detradius,
            head_length=0.3 * detradius,
            fc=detcolor,
            ec="none",
            length_includes_head=True,
        )

        # Compute the font size to use for detector labels
        fontpix = 0.1 * detradius * hpixperdeg
        if fontpix < 1.0:
            fontpix = 1.0

        if show_centers:
            ysgn = -1.0
            if dx < 0.0:
                ysgn = 1.0
            ax.text(
                (xpos + 0.1 * dx),
                (ypos + 0.1 * ysgn * dy),
                f"({xpos:0.4f}, {ypos:0.4f})",
                color="green",
                fontsize=fontpix,
                horizontalalignment="center",
                verticalalignment="center",
                bbox=dict(fc="w", ec="none", pad=1, alpha=0.0),
            )

        if labels:
            ax.text(
                xpos,
                ypos,
                pixel,
                color="k",
                fontsize=fontpix,
                horizontalalignment="center",
                verticalalignment="center",
                bbox=dict(fc="white", ec="none", pad=0.2, alpha=1.0),
            )
            labeloff = fontpix * len(pol) / hpixperdeg
            if dy < 0:
                labeloff = -labeloff
            ax.text(
                (xtail + 1.0 * dx + labeloff),
                (ytail + 1.0 * dy),
                pol,
                color="k",
                fontsize=fontpix,
                horizontalalignment="center",
                verticalalignment="center",
                bbox=dict(fc="none", ec="none", pad=0, alpha=1.0),
            )

    # Draw a "mini" coordinate axes for reference
    shortest = min(half_width, half_height)
    xmini = -0.7 * half_width
    ymini = -0.7 * half_height
    xlen = 0.06 * shortest
    ylen = 0.06 * shortest
    mini_width = 0.005 * shortest
    mini_head_width = 3 * mini_width
    mini_head_len = 3 * mini_width
    if xieta:
        aprops = [
            (xlen, 0, "-", r"$\xi$"),
            (0, ylen, "-", r"$\eta$"),
            (-xlen, 0, "--", "Y"),
            (0, -ylen, "--", "X"),
        ]
    else:
        aprops = [
            (xlen, 0, "-", "X"),
            (0, ylen, "-", "Y"),
            (-xlen, 0, "--", r"$\eta$"),
            (0, -ylen, "--", r"$\xi$"),
        ]
    for ap in aprops:
        lx = xmini + 1.5 * ap[0]
        ly = ymini + 1.5 * ap[1]
        lw = figdpi / 200.0
        ax.arrow(
            xmini,
            ymini,
            ap[0],
            ap[1],
            width=mini_width,
            head_width=mini_head_width,
            head_length=mini_head_len,
            fc="k",
            ec="k",
            linestyle=ap[2],
            linewidth=lw,
            length_includes_head=True,
        )
        ax.text(
            lx,
            ly,
            ap[3],
            color="k",
            fontsize=int(figdpi / 10),
            horizontalalignment="center",
            verticalalignment="center",
        )

    st = f"Focalplane Looking Towards Observer {lat_elstr}"
    if xieta:
        st = f"Focalplane on Sky From Observer {lat_elstr}"
    fig.suptitle(st)

    plt.savefig(outfile)
    plt.close()
    return


class clr:
    WHITE = "\033[97m"
    PURPLE = "\033[95m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    ENDC = "\033[0m"

    def disable(self):
        self.WHITE = ""
        self.PURPLE = ""
        self.BLUE = ""
        self.GREEN = ""
        self.YELLOW = ""
        self.RED = ""
        self.ENDC = ""


def summary_text(hw):
    """Print a textual summary of a hardware configuration.

    Args:
        hw (Hardware): A hardware dictionary.

    Returns:
        None

    """
    for obj, props in hw.data.items():
        nsub = len(props)
        print(
            "{}{:<12}: {}{:5d} objects{}".format(
                clr.WHITE, obj, clr.RED, nsub, clr.ENDC
            )
        )
        if nsub <= 2000:
            line = ""
            for k in list(props.keys()):
                if (len(line) + len(k)) > 72:
                    print("    {}{}{}".format(clr.BLUE, line, clr.ENDC))
                    line = ""
                line = "{}{}, ".format(line, k)
            if len(line) > 0:
                print("    {}{}{}".format(clr.BLUE, line.rstrip(", "), clr.ENDC))
        else:
            # Too many to print!
            print("    {}(Too many to print){}".format(clr.BLUE, clr.ENDC))

    return



#!/usr/bin/env python3

# Copyright (c) 2024-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""
This script runs a transfer function simulation.

You can see the automatically generated command line options with:

    toast_so_transfer --help

Or you can dump a config file with all the default values with:

    toast_so_transfer --default_toml config.toml

This script contains just comments about what is going on.  For details about all the
options for a specific Operator, see the documentation or use the help() function from
an interactive python session.

"""

import argparse
import datetime
import os
import sys
import traceback

import numpy as np
import yaml

from astropy import units as u

# Import sotodlib.toast first, since that sets default object names
# to use in toast.
import sotodlib.toast as sotoast

import toast
import toast.ops

from toast.mpi import MPI, Comm
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .. import workflows as wrk

# Make sure pixell uses a reliable FFT engine
import pixell.fft

pixell.fft.engine = "fftw"


def reduce_input(job, otherargs, runargs, data):
    """Apply cuts and compute noise model on original data."""
    job_ops = job.operators

    wrk.create_az_intervals(job, otherargs, runargs, data)
    wrk.apply_readout_filter(job, otherargs, runargs, data)
    wrk.deconvolve_detector_timeconstant(job, otherargs, runargs, data)
    wrk.select_pointing(job, otherargs, runargs, data)

    # Preprocess data to get flags / cuts
    wrk.preprocess(job, otherargs, runargs, data)

    wrk.diff_noise_estimation(job, otherargs, runargs, data)
    wrk.noise_estimation(job, otherargs, runargs, data)

    if job_ops.demodulate.enabled:
        save_weights = job_ops.weights_radec
        data = wrk.demodulate(job, otherargs, runargs, data)
        job_ops.weights_radec = save_weights

    wrk.filter_ground(job, otherargs, runargs, data)
    wrk.filter_common_mode(job, otherargs, runargs, data)
    wrk.filter_poly1d(job, otherargs, runargs, data)
    wrk.filter_poly2d(job, otherargs, runargs, data)

    wrk.diff_noise_estimation(job, otherargs, runargs, data)
    wrk.noise_estimation(job, otherargs, runargs, data)

    wrk.processing_mask(job, otherargs, runargs, data)
    wrk.flag_sso(job, otherargs, runargs, data)

    return data


def load_observations(job, otherargs, runargs, comm):
    """Load data and compute noise model and cuts"""
    log = toast.utils.Logger.get()
    job_ops = job.operators

    group_size = wrk.reduction_group_size(job, runargs, comm)
    toast_comm = toast.Comm(world=comm, groupsize=group_size)
    data = toast.Data(comm=toast_comm)

    # Load data from any format
    wrk.load_data_hdf5(job, otherargs, runargs, data)
    wrk.load_data_context(job, otherargs, runargs, data)

    # Perform any other operations on original data before zeroing
    data = reduce_input(job, otherargs, runargs, data)

    mem = toast.utils.memreport(
        msg="(whole node)", comm=data.comm.comm_world, silent=True
    )
    log.info_rank(f"After loading data:  {mem}", data.comm.comm_world)

    # Zero out detector data
    toast.ops.Reset(detdata=[defaults.det_data])

    return data


def reduce_realization(job, otherargs, runargs, data):
    """Apply data reduction to a single realization."""
    log = toast.utils.Logger.get()
    job_ops = job.operators
    save_weights = job_ops.weights_radec

    cleanup = False
    if otherargs.simulate_demodulated:
        # We are simulating already-demodulated data.
        if not job_ops.demodulate.enabled:
            msg = "Cannot start realizations with demodulated data, "
            msg += "since demodulation is disabled."
            raise RuntimeError(msg)
        processed_data = data
    else:
        wrk.filter_hwpss(job, otherargs, runargs, data)
        if job_ops.demodulate.enabled:
            cleanup = True
            processed_data = wrk.demodulate(job, otherargs, runargs, data)
        else:
            processed_data = data

    wrk.filter_ground(job, otherargs, runargs, data)
    wrk.filter_common_mode(job, otherargs, runargs, data)
    wrk.filter_poly1d(job, otherargs, runargs, data)
    wrk.filter_poly2d(job, otherargs, runargs, data)

    if job.operators.splits.enabled:
        wrk.splits(job, otherargs, runargs, processed_data)
    else:
        wrk.mapmaker(job, otherargs, runargs, processed_data)
        wrk.mapmaker_filterbin(job, otherargs, runargs, processed_data)

    # Clean up demodulated data copy if we created it.
    if cleanup:
        processed_data.close()
    del processed_data

    job_ops.weights_radec = save_weights


def signal_realizations(job, otherargs, runargs, data):
    """Run minimal reduction on signal realizations for transfer functions."""
    log = toast.utils.Logger.get()
    job_ops = job.operators

    # Determine the binner being used
    if job_ops.mapmaker.enabled:
        if job_ops.binner_final.enabled:
            map_binner = job_ops.binner_final
        else:
            map_binner = job_ops.binner
    elif job_ops.filterbin.enabled:
        map_binner = job_ops.binner
    else:
        msg = "No mapmaker is enabled!"
        raise RuntimeError(msg)

    if not map_binner.full_pointing:
        log.warning_rank(
            "Mapmaking binner not using full detector pointing.  Overriding.",
            comm=data.comm.comm_world,
        )
        map_binner.full_pointing = True

    if (
        (job_ops.demodulate.enabled and job_ops.demodulate.purge)
        and not otherargs.simulate_demodulated
    ):
        log.warning_rank(
            "Demodulation configured to purge original data.  Overriding.",
            comm=data.comm.comm_world,
        )
        job_ops.demodulate.purge = False

    # Get the file for each realization
    realizations = dict()
    if data.comm.world_rank == 0:
        with open(otherargs.realizations, "r") as f:
            realizations = yaml.safe_load(f)
    if data.comm.comm_world is not None:
        realizations = data.comm.comm_world.bcast(realizations, root=0)

    # Get the base output directory
    base_output = str(otherargs.out_dir)

    # Create the pixel distribution prior to looping over realizations.
    # This might have been created by the processing chain on the
    # original data.  We clear it and recreate it with known options,
    # in particular saving all detector pointing.  Same with any detector
    # pointing generated during the initial processing on the input data.
    if map_binner.pixel_dist in data:
        del data[map_binner.pixel_dist]
    toast.ops.Delete(
        detdata=[
            map_binner.pixel_pointing.pixels,
            map_binner.stokes_weights.weights,
            map_binner.pixel_pointing.detector_pointing.quats,
        ],
    ).apply(data)

    pix_dist = toast.ops.BuildPixelDistribution(
        pixel_dist=map_binner.pixel_dist,
        pixel_pointing=map_binner.pixel_pointing,
        save_pointing=True,
    )
    pix_dist.apply(data)
    map_binner.stokes_weights.apply(data)

    for realization_indx in list(sorted(realizations.keys())):
        realization_file = realizations[realization_indx]
        msg = f"Realization {realization_indx:04d} using {realization_file}"
        log.info_rank(msg, comm=data.comm.comm_world)

        # Replace detector data with signal
        sig_scan = toast.ops.ScanHealpixMap(
            name="sig_scan",
            file=realization_file,
            zero=True,
            save_pointing=True,
            pixel_pointing=map_binner.pixel_pointing,
            stokes_weights=map_binner.stokes_weights,
        )
        sig_scan.apply(data)

        # Set the output directory for this realization
        realization_dir = os.path.join(
            base_output,
            f"{realization_indx:04d}",
        )
        otherargs.out_dir = realization_dir

        if data.comm.world_rank == 0:
            if not os.path.isdir(otherargs.out_dir):
                os.makedirs(otherargs.out_dir, exist_ok=True)
        if data.comm.comm_world is not None:
            data.comm.comm_world.barrier()

        # Run processing
        reduce_realization(job, otherargs, runargs, data)

    otherargs.out_dir = base_output

    mem = toast.utils.memreport(
        msg="(whole node)", comm=data.comm.comm_world, silent=True
    )
    log.info_rank(f"After all signal realizations:  {mem}", data.comm.comm_world)


def main():
    env = toast.utils.Environment.get()
    log = toast.utils.Logger.get()
    gt = toast.timing.GlobalTimers.get()
    gt.start("toast_so_transfer (total)")
    timer = toast.timing.Timer()
    timer.start()

    # Get optional MPI parameters
    comm, procs, rank = toast.get_world()

    # If the user has not told us to use multiple threads,
    # then just use one.

    if "OMP_NUM_THREADS" in os.environ:
        nthread = os.environ["OMP_NUM_THREADS"]
    else:
        nthread = 1
    log.info_rank(
        f"Executing workflow with {procs} MPI tasks, each with "
        f"{nthread} OpenMP threads at {datetime.datetime.now()}",
        comm,
    )

    mem = toast.utils.memreport(msg="(whole node)", comm=comm, silent=True)
    log.info_rank(f"Start of the workflow:  {mem}", comm)

    # Argument parsing
    parser = argparse.ArgumentParser(description="SO simulation pipeline")

    parser.add_argument(
        "--out_dir",
        required=False,
        type=str,
        default="out_sat",
        help="The output directory",
    )
    parser.add_argument(
        "--obsmaps",
        required=False,
        default=False,
        action="store_true",
        help="Map each observation separately.",
    )
    parser.add_argument(
        "--simulate_demodulated",
        required=False,
        default=False,
        action="store_true",
        help="Start each realization with demodulated / downsampled data",
    )
    parser.add_argument(
        "--realizations",
        required=False,
        default=None,
        help="YAML file containing signal realizations to process",
    )
    parser.add_argument(
        "--preprocess_copy",
        required=False,
        default=False,
        action="store_true",
        help="Perform preprocessing on a copy of the data.",
    )
    parser.add_argument(
        "--log_config",
        required=False,
        default=None,
        help="Dump out config log to this yaml file",
    )

    # FIXME:  Add an extra option here to specify a file of additional
    # per observation weight factors.

    # The operators and templates we want to configure from the command line
    # or a parameter file.

    operators = list()
    templates = list()

    # Data loading
    wrk.setup_load_data_hdf5(operators)
    wrk.setup_load_data_books(operators)
    wrk.setup_load_data_context(operators)
    wrk.setup_az_intervals(operators)
    wrk.setup_diff_noise_estimation(operators)
    wrk.setup_readout_filter(operators)
    wrk.setup_deconvolve_detector_timeconstant(operators)
    wrk.setup_pointing(operators)

    # Processing that might be used.
    wrk.setup_preprocess(parser, operators)

    wrk.setup_filter_common_mode(operators)
    wrk.setup_filter_ground(operators)
    wrk.setup_filter_poly1d(operators)
    wrk.setup_filter_poly2d(operators)
    wrk.setup_noise_estimation(operators)

    wrk.setup_demodulate(operators)

    wrk.setup_processing_mask(operators)
    wrk.setup_flag_sso(operators)

    wrk.setup_splits(operators)

    wrk.setup_mapmaker(operators, templates)
    wrk.setup_mapmaker_filterbin(operators)

    job, config, otherargs, runargs = wrk.setup_job(
        parser=parser, operators=operators, templates=templates
    )

    # Create our output directory
    if comm is None or comm.rank == 0:
        if not os.path.isdir(otherargs.out_dir):
            os.makedirs(otherargs.out_dir, exist_ok=True)

    # Log the config that was actually used at runtime.
    if otherargs.log_config is not None:
        toast.config.dump_yaml(otherargs.log_config, config, comm=comm)

    # If this is a dry run, exit
    if otherargs.dry_run:
        log.info_rank("Dry-run complete", comm=comm)
        return

    # Load data and generate any noise models, cuts, and flags
    data = load_observations(job, otherargs, runargs, comm)

    # Loop over signal realizations
    signal_realizations(job, otherargs, runargs, data)

    # Collect optional timing information
    alltimers = toast.timing.gather_timers(comm=comm)
    if data.comm.world_rank == 0:
        out = os.path.join(otherargs.out_dir, "timing")
        toast.timing.dump(alltimers, out)

    log.info_rank("Workflow completed in", comm=comm, timer=timer)


def cli():
    world, procs, rank = toast.mpi.get_world()
    with toast.mpi.exception_guard(comm=world):
        main()


if __name__ == "__main__":
    cli()



"""
This submodule intends to produce a obslist dictionary that will be feed to
a mapmaker. Every element/key of obslist will be mapped as an individual
map. For example, we could request depth-1 maps which are continous scans
one after the other that will be mapped together, we could request per-obs
which will be individual observations mapped individually, etc. See the main
interface function build_obslists for more information and definitions. An
obslist is a dictionary where each element is a list of
(obs_id,detset,band,index) to be mapped. A detset is a wafer_slot by default,
unless per_tube=True in which case the detset will be over the entire
optical tube, meaning we will loop over all available wafer_slots.
Index maps to the list of individual obs. For example, an atomic map will be
a single id, a depth-1 map will be multiple ids. The keys of the obslist will
be (pid,detset,band). Here pid maps into the periods list, which is simply
the periods ranges that will be mapped. For example, for atomic maps these
periods will be simply the ctimes dividing individual obs, for depth-1 maps
it will be the ctimes dividing depth-1 maps, etc.
"""

__all__ = ['build_obslists','NoTODFound']
import numpy as np
from pixell import utils
from scipy import ndimage

from .utils import get_ids

class NoTODFound(Exception):
    def __init__(self, msg):
        self.msg = msg

def build_obslists(context, query, mode=None, nset=None, wafer=None,
        freq=None, ntod=None, tods=None, fixed_time=None, mindur=None, per_tube=False ):
    """ 
    Return an obslists dictionary (described in the submodule docstring), 
    along with all ancillary data necessary for the mapmaker
    
    Parameters
    ----------
    context : sotodlib.core.Context
            A context object
    query : str
            A list of tods. Can be a file with one obs_id per line, or a
            query that context.obsdb.query() will understand
    mode : str or none, optional
            Optional, mode for selecting tods. Can be 'per_obs', 
            'fixed_interval', 'depth_1'. Default is 'per_obs'
    per_tube : bool, optional
            If True, depth-1 maps will be returned over tube rather than
            per wafer. Overrides nset and wafer
    nset : int or None, optional
            Optional, the first nset sets will be mapped
    wafer : str or None, optional
            Specific wafer to retrieve, overrides nset
    freq : str or None, optional
            Retrieve a specific frequency bandpass
    ntod : int or None, optional
            Optional, the first ntod observations listed in query will be 
            mapped
    tods : str or None, optional
            Optional, a string with specific tods that will be mapped
    fixed_time : int or None, optional
            Optional, if mode=='fixed_interval', this is the fixed time in
            seconds
    mindur : int or None, optional
            Optional, minimum duration of an observation to be included in 
            the mapping. If not defined it will be 120 seconds
           
    Returns
    -------
    obslists : dict
        A dict that contains a list of observations that will be mapped in 
        each individual map (atomic, depth-1, etc.)
    obskeys : list
        A list of the keys of the obslists dict.
    periods : list
        A list with the ctimes edges for the individual maps (atomic, 
        depth-1, etc.)
    obs_infos : list
        A list where each element is another list of the obs_info for each 
        observation
    """
    
    # Get the full list of tods we will work with
    ids = get_ids(query, context=context)
    # restrict tod selection further. 
    #E.g. --tods [0], --tods[:1], --tods[::100], --tods[[0,1,5,10]], etc.
    if tods: ids = np.array(eval("ids"+tods)).reshape(-1)
    if ntod: ids = ids[:ntod]
    if len(ids) == 0:
        raise NoTODFound("No tods found!")
    # Extract info about the selected ids
    obs_infos = context.obsdb.query("obs_id in (%s)" % ",".join(["'%s'" % id for id in ids]))
    obs_infos = obs_infos.asarray().view(np.recarray)
    ids       = obs_infos.obs_id # can't rely on ordering, so reget
    
    if mode is None or mode == 'per_obs':
        # We simply need to make and obslists dict with each key being one obs
        periods = find_scan_periods_perobs(obs_infos)
    elif mode=='depth_1':
        periods   = find_scan_periods(obs_infos, ttol=12*3600)
        periods   = split_periods(periods, 24*3600)
    elif mode=='fixed_interval':
        if fixed_time is not None:
            periods   = find_scan_periods(obs_infos, ttol=12*3600)
            periods   = split_periods(periods, fixed_time)
        else:
            periods   = find_scan_periods(obs_infos, ttol=12*3600)
            periods   = split_periods(periods, 24*3600) # If fixed_time was not set, 
            #then we do 24 hrs by default and it will be the same as depth_1
    else:
        raise NoTODFound("Invalid mode!")
    
    # We will make one map per period-detset-band if per_tube=False
    # or per period-band if per_tube=True
    obslists = build_period_obslists(obs_infos, periods, context, nset=nset,
                                     wafer=wafer, freq=freq, per_tube=per_tube)
    obskeys  = sorted(obslists.keys())
    return obslists, obskeys, periods, obs_infos


def find_scan_periods(obs_info, ttol=60, atol=2*utils.degree, mindur=120):
    """Given a list of obs info, return the set of contiguous scanning periods in the form
    [:,{ctime_from,ctime_to}]."""
    atol = atol/utils.degree
    info = np.array([obs_info[a] for a in ["az_center", "el_center", 
                                           "az_throw", "timestamp", "duration"]]).T
    # Get rid of nan entries
    bad  = np.any(~np.isfinite(info),1)
    # get rid of too short tods, since those don't have reliable az bounds
    bad |= info[:,-1] < mindur
    info = info[~bad]
    t1   = info[:,-2] # timestamp is start of tod, unlike in enki
    info = info[np.argsort(t1)]
    # Start, end
    t1   = info[:,-2]
    t2   = t1 + info[:,-1]
    # Remove angle ambiguities
    info[:,0] = utils.rewind(info[:,0], period=360)
    # How to find jumps:
    # 1. It's a jump if the scanning changes
    # 2. It's also a jump if a the interval between tod-ends and tod-starts becomes too big
    changes    = np.abs(info[1:,:3]-info[:-1,:3])
    jumps      = np.any(changes > atol,1)
    jumps      = np.concatenate([[0], jumps]) # from diff-inds to normal inds
    # Time in the middle of each gap
    gap_times = np.mean(find_period_gaps(np.array([t1,t2]).T, ttol=ttol),1)
    gap_inds  = np.searchsorted(t1, gap_times)
    jumps[gap_inds] = True
    labels  = np.cumsum(jumps)
    linds   = np.arange(np.max(labels)+1)
    t1s     = ndimage.minimum(t1, labels, linds)
    t2s     = ndimage.maximum(t2, labels, linds)
    # Periods is [nperiod,{start,end}] in ctime. Start is the start of the first tod
    # in the scanning period. End is the end of the last tod in the scanning period.
    periods = np.array([t1s, t2s]).T
    return periods

def find_scan_periods_perobs(obs_info, mindur=120):
    """Given a list of obs info, return the periods per obs, i.e. start and 
    stop of each scan. This is a simplified version of find_scan_periods
    [:,{ctime_from,ctime_to}]."""
    info = np.array([obs_info[a] for a in ["timestamp", "duration"]]).T
    # Get rid of nan entries
    bad  = np.any(~np.isfinite(info),1)
    # get rid of too short tods, since those don't have reliable az bounds
    bad |= info[:,-1] < mindur
    info = info[~bad]
    t1   = info[:,-2] # timestamp is start of tod, unlike in enki
    info = info[np.argsort(t1)]
    # Start, end
    t1   = info[:,-2]
    t2   = t1 + info[:,-1]
    periods = np.array([t1, t2]).T
    return periods

def find_period_gaps(periods, ttol=60):
    """Helper for find_scan_periods. Given the [:,{ctime_from,ctime_to}] for all
    the individual scans, returns the times at which the gap between the end of
    a tod and the start of the next is greater than ttol (default 60 seconds)."""
    # We want to sort these and look for any places
    # where a to is followed by a from too far away. To do this we need to keep
    # track of which entries in the combined, sorted array was a from or a to
    periods = np.asarray(periods)
    types   = np.zeros(periods.shape, int)
    types[:,1] = 1
    types   = types.reshape(-1)
    ts      = periods.reshape(-1)
    order   = np.argsort(ts)
    ts, types = ts[order], types[order]
    # Now look for jumps
    jumps = np.where((ts[1:]-ts[:-1] > ttol) & (types[1:]-types[:-1] < 0))[0]
    # We will return the time corresponding to each gap
    gap_times = np.array([ts[jumps], ts[jumps+1]]).T
    return gap_times

def split_periods(periods, maxdur):
    # How long is each period
    durs   = periods[:,1]-periods[:,0]
    # How many parts to split each into
    nsplit = utils.ceil(durs/maxdur)
    nout   = np.sum(nsplit)
    # For each output period, find which input period it
    # corresponds to
    group  = np.repeat(np.arange(len(durs)), nsplit)
    sub    = np.arange(nout)-np.repeat(utils.cumsum(nsplit),nsplit)
    t1     = periods[group,0] + sub*maxdur
    t2     = np.minimum(periods[group,0]+(sub+1)*maxdur, periods[group,1])
    return np.array([t1,t2]).T

def build_period_obslists(obs_info, periods, context, nset=None, 
                          wafer=None, freq=None, per_tube=False):
    """
    For each period for each detset-band, make a list of (id,detset,band)
    for the ids that fall inside that period. Returns the obslist dictionary
    that maps (pid,deset,band) to those lists. pid is here the index into
    the periods, where periods is [nperiod,{ctime_from,ctime_to}].
    If per_tube=True, then we loop over all available wafers (which will
    override nset and wafer).
    """
    obslists = {}
    # 1. Figure out which period each obs belongs to
    ctimes_mid = obs_info.timestamp + obs_info.duration/2
    pids       = np.searchsorted(periods[:,0], ctimes_mid)-1
    # 2. Build our lists. Not sure how to do this without looping
    for i, row in enumerate(obs_info):
        if wafer is not None and not per_tube:
            wafer_list = [wafer]
        else:
            wafer_list = row.wafer_slots_list.split(',')
        if freq is not None:
            band_list = [freq]
        else:
            if row.tube_flavor in ['mf','MF']:
                band_list = ['f090', 'f150']
            elif row.tube_flavor in ['uhf','UHF']:
                band_list = ['f220', 'f280']
            elif row.tube_flavor in ['lf','LF']:
                raise ValueError('Band list for lf not implemented yet.')
        if per_tube:
            for band in band_list:
                key = (pids[i], row.tube_slot, band)
                if key not in obslists: obslists[key] = []
                for detset in wafer_list:
                    obslists[key].append((row.obs_id, detset, band, i))
        else:
            for detset in wafer_list[:nset]:
                for band in band_list:
                    key = (pids[i], detset, band)
                    if key not in obslists: obslists[key] = []
                    obslists[key].append((row.obs_id, detset, band, i))
    return obslists



# Copyright (c) 2018-2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Metadata containers.
"""

from .resultset import ResultSet
from .detdb import DetDb
from .obsdb import ObsDb
from .obsfiledb import ObsFileDb
from .manifest import ManifestDb, ManifestScheme
from .loader import (
    SuperLoader, LoaderInterface, MetadataSpec,
    merge_det_info, load_metadata,
)
from . import cli

def get_example(db_type, *args, **kwargs):
    if db_type == 'DetDb':
        from .detdb import get_example
        return get_example(*args, **kwargs)
    else:
        raise ValueError('Unknown db_type: %s' % db_type)



import so3g.proj
import numpy as np
from pixell import enmap, wcsutils, utils
from . import healpix_utils as hp_utils

import time
import re

DEG = np.pi/180


def _get_csl(sight):
    """Return the CelestialSightLine equivalent of sight.  If sight is
    already of that class, it is returned.  If it's a G3VectorQuat or
    an array [n,4], those coordinates are wrapped and returned (note a
    reference may be taken rather than a copy made).

    """
    if isinstance(sight, so3g.proj.CelestialSightLine):
        return sight
    if isinstance(sight, so3g.proj.quat.G3VectorQuat):
        c = so3g.proj.CelestialSightLine()
        c.Q = sight
        return c
    qa = np.asarray(sight, dtype='float')
    c = so3g.proj.CelestialSightLine()
    c.Q = so3g.proj.quat.G3VectorQuat(qa)
    return c

def _valid_arg(*args, src=None):
    """Return some data, possibly extracted from an AxisManager (or
    dict...), based on the arguments args.  This is to help with
    processing function arguments that override a default behavior
    which is to look up a thing in axisman.  For example::

      signal = _valid_arg(signal, 'signal', src=axisman)

    is equivalent to::

      if signal is None:
          signal = 'signal'
      if isinstance(signal, str):
          signal = axisman[signal]

    Similarly::

        sight = _valid_arg(sight, src=axisman)

    is a shorthand for::

        if sight is not None:
           if isinstance(sight, string):
               sight = axisman[sight]

    Each element of args should be either a data vector (non string),
    a string, or None.  The arguments are processed in order.  The
    first argument that is not None will cause the function to return;
    if that argument k is a string, then axisman[k] is returned;
    otherwise k is returned directly.  If all arguments are None, then
    None is returned.

    """
    for a in args:
        if a is not None:
            if isinstance(a, str):
                try:
                    a = src[a]
                except TypeError:
                    raise TypeError(f"Tried to look up '{a}' in axisman={axisman}")
            return a
    return None

def _not_both(a, b, name='{item}', dtype=None):
    if a is not None:
        if b is not None:
            raise ValueError('self.%s and kwarg %s both not None!' % (name, name))
        b = a
    if b is not None and dtype is not None:
        b = b.astype(dtype)
    return b

class Timer:
    """Context manager that prints elapsed time to terminal or a log or
    whatever.  For example::

        with Timer('tod-to-map projection', logger.info):
            P.to_map(tod)

    Should log a message of the form::

        INFO:tod-to-map projection:            3.212 seconds

    Use the fmt keyword argument to enter your own format string,
    which can include references to {msg}, {start_time}, {end_time},
    {elapsed}.

    To avoid repeating the same arguments, subclass this and set
    FORMAT and PRINT_FUNC for your use cases::

       class MyTimer(Timer):
           FORMAT = 'mapmaker7: {msg} took {elapsed:10.3f} seconds'
           PRINT_FUNC = lambda x: (logger.info(x), print(x))

    """
    FMT = '{msg:40}: {elapsed:10.3f} seconds'
    PRINT_FUNC = print

    def __init__(self, msg=None, print_func=None, fmt=None):
        if msg is None:
            msg = 'timed operation'
        if fmt is None:
            fmt = self.FMT
        if print_func is None:
            print_func = self.PRINT_FUNC
        self.msg = msg
        self.fmt = fmt
        self.print_func = print_func
        self.start_time = time.time()
    def __enter__(self):
        return self
    def __exit__(self, *args, **kw):
        self.end_time = time.time()
        self.elapsed = self.end_time - self.start_time
        self.text = self.fmt.format(
            msg=self.msg, start_time=self.start_time, end_time=self.end_time,
            elapsed=self.elapsed)
        self.print_func(self.text)

def get_fplane(tod, dets=None, focal_plane=None, hwp=False):
    """Return a FocalPlane object for the AxisManager tod.

    Args:
      dets (list of str): If set, the detectors will be restricted to this list.
      focal_plane (AxisManager): A sub-axis-manager containing the detector
        layout and response information. Defaults to tod.focal_plane.
        This must contain at least the xi and eta members. Optional members are
        gamma, T, P, Q and U, which are used to specify the detector response.
        Typically these would be used like this:
          gamma: Only detector polariztion angle specified. Detectors assumed to
            have 100% response in total intensity and polarization.
          gamma, P: Detector polarization angle and polarization efficiency.
          gamma, T, P: As above, but also including total intensity. This is
            usually unnecessary, but in demodulation mapmaking it's sometimes
            useful to support polarization-sensitive detectors with no sensitivity
            to total intensity.
          T, Q, U: Alternative to specifying polarization angles. The detectors'
            response to T, Q and U are directly specified in the focal plane frame.
            Equivalent to gamma = arctan(U,Q)/2, P = (Q**2+U**2)**0.5.

    Returns:
      A FocalPlane object, which simply encapsulates a numpy array of detector
      quaternions and a numpy array of detector responses to total intensity
      and polarization.
    """
    dets   = _valid_arg(dets, tod.dets.vals, src=tod)
    fp     = _valid_arg(focal_plane, 'focal_plane', src=tod)
    fp     = fp.restrict("dets", dets)
    kwargs = {key:getattr(fp,key) for key in ["gamma","T","P","Q","U"] if key in fp}
    fplane = so3g.proj.FocalPlane.from_xieta(fp.xi, fp.eta, **kwargs, hwp=hwp)
    return fplane

def get_radec(tod, wrap=False, dets=None, timestamps=None, focal_plane=None,
              boresight=None, sight=None):
    """Get the celestial coordinates of all detectors at all times.

    Args:
      wrap (bool): If True, the output is stored into tod['radec'].
        This can also be a string, in which case the result is stored
        in tod[wrap].
      dets (list of str): If set, then coordinates are only computed
        for the requested detectors.  Note you probably can't wrap the
        result, in this case, as the dets axis is non-concordant.

    Returns:
      The returned array has shape (n_det, n_samp, 4).  The four
      components in the last dimension correspond to (lon, lat,
      cos(psi), sin(psi)).  The lon and lat are in radians and
      correspond to RA and dec of equatorial coordinates.  Psi is is
      the parallactic rotation, measured from North towards West
      (opposite the direction of standard Position Angle).
    """
    dets = _valid_arg(dets, tod.dets.vals, src=tod)
    if sight is None:
        timestamps = _valid_arg(timestamps, 'timestamps', src=tod)
        boresight = _valid_arg(boresight, 'boresight', src=tod)
        sight = so3g.proj.CelestialSightLine.az_el(
            timestamps, boresight.az, boresight.el, roll=boresight.roll,
            site='so', weather='typical')
    else:
        sight = _get_csl(_valid_arg(sight, 'sight', src=tod))
    fplane = get_fplane(tod, dets=dets, focal_plane=focal_plane)
    asm    = so3g.proj.Assembly.attach(sight, fplane)
    output = np.zeros((len(dets), len(sight.Q), 4))
    proj = so3g.proj.Projectionist()
    proj.get_coords(asm, output=output)
    if wrap is True:
        wrap = 'radec'
    if wrap:
        tod.wrap(wrap, output, [(0, 'dets'), (1, 'samps')])
    return output

def get_horiz(tod, wrap=False, dets=None, timestamps=None, focal_plane=None,
              boresight=None):
    """Get the horizon coordinates of all detectors at all times.

    Args:
      wrap (bool): If True, the output is stored into tod['horiz'].
        This can also be a string, in which case the result is stored
        in tod[wrap].
      dets (list of str): If set, then coordinates are only computed
        for the requested detectors.  Note you probably can't wrap the
        result, in this case, as the dets axis is non-concordant.

    Returns:
      The returned array has shape (n_det, n_samp, 4).  The four
      components in the last dimension correspond to (-lon, lat,
      cos(psi), sin(psi)).  The -lon and lat are in radians and
      correspond to horizon azimuth and elevation.  Psi is is
      the parallactic rotation, measured from North towards West
      (opposite the direction of standard Position Angle).

    """
    dets = _valid_arg(dets, tod.dets.vals, src=tod)
    timestamps = _valid_arg(timestamps, 'timestamps', src=tod)
    boresight = _valid_arg(boresight, 'boresight', src=tod)

    sight = so3g.proj.CelestialSightLine.for_horizon(
        timestamps, boresight.az, boresight.el, roll=boresight.roll)

    fplane = get_fplane(tod, dets=dets, focal_plane=focal_plane)
    asm = so3g.proj.Assembly.attach(sight, fplane)
    output = np.zeros((len(dets), len(timestamps), 4))
    proj = so3g.proj.Projectionist()
    proj.get_coords(asm, output=output)
    # The lonlat pair is (-az, el), so restore the az sign.
    output[:,:,0] *= -1
    if wrap is True:
        wrap = 'horiz'
    if wrap:
        tod.wrap(wrap, output, [(0, 'dets'), (1, 'samps')])
    return output


def get_wcs_kernel(proj, ra=None, dec=None, res=None):
    """Construct a WCS.  This fixes the projection type (e.g. CAR, TAN),
    centered with respect to a reference point (at ra,dec = (0,0)), and
    resolution of a pixelization, without specifying a particular grid
    of pixels.

    This interface is subject to change.

    Args:
      proj (str): Either the name of a FITS projection to use (e.g. 'car',
        'cea', 'tan'), in which case "res" must also be specified, or a
        string containing both the projection and the resolution, in this
        format:

                      proj-res

        where proj is a projection type (e.g. 'car', 'tan, 'gnom') and res
        is the resolution, in appropriate units (deg, arcmin or arcsec).
        Examples of acceptable args:

                      'car-0.5deg'
                      'tan-3arcmin'
                      'gnom-25arcsec'

      ra: Right Ascension (longitude) of the reference position, in
        radians.
      dec: Declination (latitude) of the reference position, in
        radians.
      res: Resolution, in radians.

    Returns a WCS object that captures the requested pixelization.

    """
    if len(proj) > 4:
        m = re.match('(?P<proj>car|tan|cea|gnom)-(?P<res>[0-9]*.?[0-9]*)(?P<unit>arcsec|arcmin|deg)',
                     proj)
        if m is None:
            raise ValueError("Input projection string is incorrectly formatted.")
        proj = m['proj']
        res = float(m['res'])
        unit = m['unit']

        # Convert to radians
        res = res * np.pi/180.
        if unit == 'arcmin':
            res /= 60.
        if unit == 'arcsec':
            res /= 3600.

        _, wcs = enmap.geometry(np.array((0, 0)), shape=(1, 1), proj=proj,
                                res=(res, -res))

    else:
        assert np.isscalar(res)  # This ain't enlib.
        _, wcs = enmap.geometry(np.array((dec, ra)), shape=(1, 1), proj=proj,
                                res=(res, -res))

    if proj == 'car':
        wcs.wcs.crpix = [1.0, 0.5]

        # Check if the resolution is sensible, for SHTs
        res = res * 180./np.pi  # Easier to do this check in degs
        div = 90./res - np.round(90./res)
        if abs(div) > 1e-8:
            raise ValueError(f"The requested resolution {res} deg does not divide 90 deg evenly.")

    return wcs


def get_footprint(tod, wcs_kernel, dets=None, timestamps=None, boresight=None,
                  focal_plane=None, sight=None, rot=None):
    """Find a geometry (in the sense of enmap) based on wcs_kernel that is
    big enough to contain all data from tod.  Returns (shape, wcs).

    If wcs_kernel is a string in the format "NSIDE=1024", returns a healpix geometry object
    for an un-tiled Healpix map at given NSIDE with default ordering.

    """
    if isinstance(wcs_kernel, str) and "nside" in wcs_kernel.lower():
        wcs_kernel = wcs_kernel.lower().strip().replace(' ', '')
        # Assume we are in the format "nside=1024"
        nside = int(wcs_kernel[6:])
        return hp_utils.get_geometry(nside)

    if isinstance(wcs_kernel, str):
        wcs_kernel = get_wcs_kernel(wcs_kernel)

    dets = _valid_arg(dets, tod.dets.vals, src=tod)
    fp0 = _valid_arg(focal_plane, 'focal_plane', src=tod)
    if sight is None and 'sight' in tod:
        sight = tod.sight
    sight = _valid_arg(sight, tod.get('sight'), src=tod)
    if sight is None:
        # Let's try to either require a sightline or boresight info.
        timestamps = _valid_arg(timestamps, 'timestamps', src=tod)
        boresight = _valid_arg(boresight, 'boresight', src=tod)
        sight = so3g.proj.CelestialSightLine.az_el(
            timestamps, boresight.az, boresight.el, roll=boresight.roll,
            site='so', weather='typical').Q
    sight = _get_csl(sight)
    n_samp = len(sight.Q)

    # Get a convex hull focal plane.
    (xi0, eta0), R, xieta1 = get_focal_plane_cover(
        focal_plane=fp0, count=16)
    fp1 = so3g.proj.FocalPlane.from_xieta(xieta1[0], xieta1[1])

    asm    = so3g.proj.Assembly.attach(sight, fp1)
    planar = np.zeros((xieta1.shape[1], n_samp, 4))
    proj   = so3g.proj.Projectionist.for_geom((1,1), wcs_kernel)
    if rot:
        # Works whether rot is a quat or a vector of them.
        asm.Q = rot * asm.Q
    proj.get_planar(asm, output=planar)
    # planar is now [ndet,nsamp,{ira,idec}] in intermediate
    # coordinates in radians. These will be 0 at the
    # reference point, and span at most [-pi,pi] in ira
    # and [-pi/2,pi/2] in idec

    # Sometimes a patch will end up straddling the edege of the
    # map. For example, for a standard CAR projection with a
    # reference point at ra=dec=0, a patch centered on ra=180
    # would be half on one side of the map, half on the other,
    # with a wide stretch of nothing between. For most projections
    # there's nothing we can do about this, but for the special case
    # of non-oblique cylindrical projections, we can modify crval
    # and crpix to construct a compatible pixelization centered on
    # the patch center.
    recenter = wcsutils.is_separable(wcs_kernel)
    if recenter:
        # First part of recentering:
        # Unwind each detector in ra to avoid angle jumps, letting
        # us measure the actual extent of the patch.
        planar[:,:,0] = utils.unwind(planar[:,:,0])
        # Harmonize detectors. This assumes that the detectors won't
        # be more than 180° away from each other at any given time
        offs = np.round((planar[:,0,0]-planar[0,0,0])/(2*np.pi))*(2*np.pi)
        planar[:,:,0] -= offs[:,None]

    # Go from intermediate coordinates to pixel coordiantes.
    # We add crpix because planar is relative to the ref-point.
    # NB! This makes it 1-based!
    delts   = wcs_kernel.wcs.cdelt * DEG
    pixbox  = utils.minmax(planar[:,:,:2]/delts,(0,1))
    pixbox += wcs_kernel.wcs.crpix
    del planar
    # Use these to construct a wcs with lower-left corner
    # as close to pixbox[0] as possible. We don't want to
    # change the pixel grid alignment though, so round to
    # nearest whole pixel
    p1, p2 = utils.floor(pixbox)
    w = wcs_kernel.deepcopy()
    # Adjust crpix so p1 → 1.
    w.wcs.crpix -= p1-1
    # If we cover the whole width of the sky, even after unwrapping,
    # then this might end up being 1 pixel too wide, giving a sky
    # slightly > 360° which can cause some minor problems.
    # Can't remove the +1 though, as that would sometimes chop off
    # a pixel from the exposed area. Either just ignore, since
    # this is a hypothetical case very unlikely to actually happen,
    # or add a special case. NB! This case *will* trigger regularly
    # due to wrapping if recentering is turned off.
    shape = (p2-p1+1)[::-1]

    if recenter:
        # Second part of recentering. Modify crval[0]
        x_mid  = shape[-1]//2+1
        ra_mid = w.wcs.crval[0] + (x_mid-w.wcs.crpix[0])*w.wcs.cdelt[0]
        w.wcs.crpix[0] = x_mid
        w.wcs.crval[0] = ra_mid

    return (shape, w)


def get_focal_plane_cover(tod=None, count=0, focal_plane=None,
                          xieta=None, det_weights=None):
    """Process a bunch of detector positions into a center and radius such
    that a circle with that center and radius contains all the
    detectors.  Also return detector positions, arranged approximately
    in a circle, whose convex hull contains all input detectors.

    Args:
      tod (AxisManager): source of focal_plane, if it is not passed.
      count (int): number of points on the enclosing circle to return.
      focal_plane (AxisManager): source for xi and eta, if xieta is
        not passed in
      xieta (array): (2, n) array (or similar) of xi and eta
        detector positions.
      det_weights (array): If provided, must be same length as the xi
        and eta vectors. Only dets with non-zero value for det_weights
        will be included in the evaluation of the cover.

    Returns:
      xieta0: array[2] with array center, (xi0, eta0).
      radius: radius of the enclosing circle.
      xietas: array with shape (2, count) carrying the xi and eta
        coords of the circular convex hull.

    Notes:
      If count=0, an empty list is returned for xietas. Otherwise,
      count must be at least 3 so that the shape is not degenerate.

      Any xi, eta that are not finite (e.g. nan or inf) are excluded
      from the computation. If no detectors remain after the combined
      finiteness and det_weights cuts, a ValueError is raised.

      Note that ``det_weights`` can be int, float, or bool
      type. Sometimes you might want to only include optical dets in
      the result; e.g.::

          ..., det_weights=(aman.det_info.wafer.type=="OPTC"), ...

      In degenerate cases (all dets are in exactly the same place), a
      radius of zero may be returned.

    """
    if xieta is None:
        if focal_plane is None:
            focal_plane = tod.focal_plane
        xi = focal_plane.xi
        eta = focal_plane.eta
    else:
        xi, eta = xieta[:2]
    mask = np.isfinite(xi) * np.isfinite(eta)
    if det_weights is not None:
        mask *= det_weights.astype(bool)

    if not np.any(mask):
        raise ValueError('All provided (xi, eta) coords are excluded; '
                         'cannot estimate a focal plane cover.')

    # Restrict to only dets under consideration.
    xi, eta = xi[mask], eta[mask]

    qs = so3g.proj.quat.rotation_xieta(xi, eta)

    # Starting guess for center
    xi0, eta0 = xi.mean(), eta.mean()
    for i in range(10):
        q = so3g.proj.quat.rotation_xieta(xi0, eta0)
        d_xi, d_eta, _ = so3g.proj.quat.decompose_xieta(~q * qs)
        if abs(d_xi.mean()) + abs(d_eta.mean()) < 1e-5*DEG:
            break
        xi0 += d_xi.mean()*.8
        eta0 += d_eta.mean()*.8
    else:
        raise ValueError('No convergence?')
    R = (d_xi**2 + d_eta**2).max()**.5
    if count == 0:
        return ((xi0, eta0), R, [])
    if count < 3:
        raise ValueError('count must be 0 or >=3.')

    dphi = 2*np.pi/count
    phi = np.arange(count) * dphi
    # cos(dphi/2) is the largest underestimate in radius one can make when
    # replacing a circle with an n_circ-sided polygon, as we do here.
    L = 1.01 * R / np.cos(dphi/2)
    xi, eta = L * np.cos(phi), L * np.sin(phi)

    # Rotate those into place.
    xi, eta, _ = so3g.proj.quat.decompose_xieta(
        q * so3g.proj.quat.rotation_xieta(xi, eta))
    return (np.array([xi0, eta0]), R, np.array([xi, eta]))

def get_supergeom(*geoms, tol=1e-3):
    """Given a set of compatible geometries [(shape0, wcs0), (shape1,
    wcs1), ...], return a geometry (shape, wcs) that includes all of
    them as a subset.

    Each argument can be a geometry (shape, wcs) or an ndmap.

    """
    def as_geom(item):
        if isinstance(item, enmap.ndmap):
            shape, wcs = item.geometry
        else:
            shape, wcs = item
        return (shape[-2:], wcs)

    s0, w0 = as_geom(geoms[0])
    w0 = w0.deepcopy()

    for item in geoms[1:]:
        s, w = as_geom(item)
        # is_compatible is necessary but not sufficient.
        if not wcsutils.is_compatible(w0, w):
            raise ValueError('Incompatible wcs: %s <- %s' % (w0, w))

        # Depending on the projection, it may be possible to translate
        # crval and crpix along each dimension and maintain exact
        # pixel center correspondence.
        translate = (False, False)
        if wcsutils.is_plain(w0):
            translate = (True, True)
        elif wcsutils.is_cyl(w0) and w0.wcs.crval[1] == 0.:
            translate = (True, False)

        cdelt = w0.wcs.cdelt
        if np.any(abs(w.wcs.cdelt - cdelt) / cdelt > tol):
            raise ValueError("CDELT not the same.")

        # Determine what shift in w.crpix would make the crval the same.
        d_crpix = w.wcs.crpix - w0.wcs.crpix
        d_crval = w.wcs.crval - w0.wcs.crval
        tweak_crpix = [0, 0]
        for axis in [0, 1]:
            if d_crval[axis] != 0 and not translate[axis]:
                raise ValueError(f"Incompatible CRVAL in axis {axis}")
            d = d_crval[axis] / cdelt[axis] - d_crpix[axis]
            if abs((d + 0.5) % 1 - 0.5) > tol:
                raise ValueError(f"CRVAL not separated by integer pix in axis {axis}.")
            tweak_crpix[axis] = int(np.round(d))

        d = np.array(tweak_crpix[::-1])
        # Position of s in w0?
        corner_a = d + [0, 0]
        corner_b = d + s
        # Super footprint, in w0.
        corner_a = np.min([corner_a, [0, 0]], axis=0)
        corner_b = np.max([corner_b, s0], axis=0)
        # Boost the WCS
        w0.wcs.crpix -= corner_a[::-1]
        s0 = corner_b - corner_a
    return tuple(map(int, s0)), w0

def _confirm_wcs(*wcss):
    """Insist that all arguments are either the same wcs, or None.

    Raises a ValueError if more than one argument is a wcs, and not
    all wcs agree.  Returns either the first valid wcs, or None if
    there aren't any.

    """
    wcs_to_use = None
    for i, wcs in enumerate(wcss):
        if wcs is None:
            continue
        if wcs_to_use is None:
            wcs_to_use = wcs
        elif not wcsutils.equal(wcs_to_use, wcs):
            raise ValueError(
                f'The wcs from {i}th item ({wcs}) is discordant with prior ones ({wcs_to_use})')
    return wcs_to_use

def _invert_weights_map(weights, eigentol=1e-6, kill_partials=True,
                        UPLO='U'):
    """Compute an inverse weights matrix, using eigendecomposition methods
    that are safe against singular matrices.  This is similar to
    scipy.linalg.pinvh, but applied to each pixel in a map in an
    efficient way.

    Args:
      weights (array): an array (or ndarray) with at least 2
        dimensions, where the leading two dimensions represent
        submatrices that are to be inverted; the other dimensions
        index "the pixel".  Valid shapes would be, for example, (3, 3,
        200, 100) or (3, 3, 10231230).
      eigentol (float): sets the threshold for keeping eigenvectors.
        In each sub-matrix inversion, any eigenvectors whose absolute
        values are less than eigentol times the largest absolute
        eigenvalue are set to zero (and thus excluded from inclusion
        in the inverse).
      kill_partials (bool): if True, then pixels where any
        eigenvectors are zero (or have been forced to zero) will have
        all their eigenvectors forced to zero.  Stated another way,
        all pixels with singular or nearly singular weights
        sub-matrices will be treated as having no weight at all.
      UPLO (str): this can be 'U' or 'L', signifying that only the
        upper diagonal or lower diagonal (respectively) elements of
        each weights sub-matrix should be considered.  (This argument
        is passed through to np.linalg.eigh.)

    Returns:
      A matrix with the same shape as weights, but where the submatrix
      carried in the first two dimensions is the inverse of the
      corresponding submatrix of weights; or possibly a pseudo-inverse
      or the zero matrix depending on arguments described above.

    """
    # Quick short circuit in trivial case...
    if weights.shape[:2] == (1, 1):
        iw = np.zeros_like(weights)  # yes, this preserves wcs
        iw[weights!=0] = 1./weights[weights!=0]
        return iw

    # Collapse and reindex weights map so it is (npix, n, n).
    w = weights.reshape(weights.shape[:2] + (-1,)).transpose(2, 0, 1)

    # Get eigendecomposition of each (n, n) sub-matrix
    v, U = np.linalg.eigh(w, UPLO)

    # Identify acceptable eigenvalues -- reject ones that are non-positive or too
    # small relative to max eigenvalue in their pixel.
    eig_ok = (v > 0) * (v > v[:,-1:] * eigentol)

    # Does one bad eig spoil the basket?
    if kill_partials:
        eig_ok *= np.all(eig_ok, axis=1)[:,None]

    # Force each unacceptable eigenmode to 0.
    U *= eig_ok[:,None,:]

    # Set bad eigenvalues to 1, to avoid the divide-by-zero.
    v[~eig_ok] = 1.

    # Compute the effective inverse, U (1/diag(v)) U.T.
    A = (U / v[:,None,:])
    B = U.transpose(0,2,1)
    iw = np.matmul(A, B)

    # Reshape the output to match what was passed in.
    return iw.transpose(1,2,0).reshape(weights.shape)

def _apply_inverse_weights_map(inverse_weights, target, out=None):
    """Apply a map of matrices to a map of vectors.

    Assumes inverse_weights.shape = (a, b, ...) and target.shape =
    (b, ...); the result has shape (a, ...).

    """
    if out is None:
        out = np.empty(inverse_weights.shape[1:],
                       dtype=target.dtype)
        if isinstance(target, enmap.ndmap):
            out = enmap.ndmap(out, target.wcs)
    # Recall matmul(a, b) operates on the last two axes of (a, b). So
    # move axes, and create a second one in target; re-order at end.
    iw = np.moveaxis(inverse_weights, (0,1), (-2,-1))
    t  = np.moveaxis(target[:,None],  (0,1), (-2,-1))
    out_moved = np.moveaxis(out[:,None], (0,1), (-2,-1))
    np.matmul(iw, t, out=out_moved)
    return out

class ScalarLastQuat(np.ndarray):
    """Wrapper class for numpy arrays carrying quaternions with the ijk1
    signature.

    The practice in many codes, including TOAST, quaternionarray, and
    scipy is to store quaternions in numpy arrays with shape (..., 4),
    with the real part of the quaternion at index [..., 3].  In
    contrast, spt3g_software inherits the 1ijk from boost.

    This class serves two main purposes:

    - It can be used to annotate numpy arrays as carrying quaternions
      in the ijk1 signature (without disturbing their behavior as
      numpy arrays).
    - It provides conversion to and from G3 quaternion containers,
      including signature correction.

    When instantiating an object of this class with a numpy array as
    an argument, the object will wrap a view of the array (not a
    copy).

    When instantiating an object of this class with a G3 quat as an
    argument, the object will store a copy of the G3 quat, translated
    to ijk1 signature::

      q_g3 = spt3g.core.quat(1.,2.,3.,4.)
      q_tq = ScalarLastQuat(q_g3)
      q_g3_again = q_tq.to_g3()

      print(q_g3, q_tq, q_g3_again)
      =>  (1,2,3,4) [2. 3. 4. 1.] (1,2,3,4)

    """
    def __new__(cls, arr):
        if isinstance(arr, so3g.proj.quat.G3VectorQuat):
            arr = np.asarray(arr)
            obj = np.empty(arr.shape)
            obj[:,:3] = arr[:,1:]
            obj[:,3] = arr[:,0]
        elif isinstance(arr, so3g.proj.quat.quat):
            obj = np.array((arr.b, arr.c, arr.d, arr.a))
        else:
            obj = np.asarray(arr)
        return obj.view(cls)

    def to_g3(self):
        """Return the ijk1-signature equivalent of the enclosed quaternion
        array, as a spt3g.core.quat (if 1-d) or a G3VectorQuat (if
        2-d).

        """
        if self.shape[-1] != 4:
            raise ValueError("Last axis must have 4 elements.")
        if self.ndim == 1:
            b, c, d, a = self[:].astype(float)
            return so3g.proj.quat.quat(a, b, c, d)
        if self.ndim == 2:
            temp = np.zeros(self.shape, float)
            temp[..., 0] = self[..., 3]
            temp[..., 1:] = self[..., :3]
            return so3g.proj.quat.G3VectorQuat(temp)
        raise ValueError("Can only convert 1- or 2-d arrays to G3.")

def get_deflected_sightline(aman, wobble_meta, site='so', weather='typical'):
    """
    Constructs a deflected CelestialSightLine using HWP-synchronous
    pointing correction using combined wobble metadata that contains 
    both amp and phase fields.
    
    This function will raise ValueError unless all detectors belong to a single
    wafer and frequency band. It extracts the corresponding deflection amplitude
    and phase from the metadata, computes the wobble correction quaternion, and
    applies it to the boresight pointing.

    Parameters
    ----------
    aman : AxisManager
        AxisManager for the observation, must include hwp_angle, timestamps, 
        and boresight.az/el, as well as det_info with wafer and band info.

    wobble_meta : AxisManager
        Metadata tree containing both amp and phase fields under
        wobble_meta.{amp, phase}

    site : str
        Observatory site identifier for sightline generation (default 'so').

    weather : str
        Atmospheric condition tag for sightline model (default 'typical').

    Returns
    -------
    sight : CelestialSightLine
        The sightline with the wobble correction quaternion applied.
    """
    wafer_slots = np.unique(aman.det_info.wafer_slot)
    bands = np.unique(aman.det_info.wafer.bandpass)

    if len(wafer_slots) != 1 or len(bands) != 1:
        raise ValueError("Detectors span multiple wafer_slots or bands.")
    # the amp and phase are the same for a given wafer, so we can take any of them, in this case for detector index 0
    # !!!!! this won't work for mixing more than one wafer.
    # the metadata has amplitudes in arcmin, and phases in radians
    amp = wobble_meta.amp[0]/60.*np.pi/180.0
    phase = wobble_meta.phase[0]

    dxi = amp * np.cos(aman.hwp_angle - phase)
    deta = -amp * np.sin(aman.hwp_angle - phase)
    deflq = so3g.proj.quat.rotation_xieta(xi=dxi, eta=deta)

    sight = so3g.proj.CelestialSightLine.az_el(
        aman.timestamps,
        aman.boresight.az,
        aman.boresight.el,
        roll=aman.boresight.roll,
        weather=weather,
        site=site,
    )
    sight.Q = sight.Q * ~deflq
    return sight




import numpy as np
from sotodlib.core import FlagManager, AxisManager

def det_splits_relative(aman, det_left_right=False, det_upper_lower=False, det_in_out=False, wrap=None):
    """
    Function for adding relative detector splits to aman. A new FlagManager called det_flags will be created and the flags put there.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    det_left_right: Bool
        Perform a detector left/right split
    det_upper_lower: Bool
        Perform a detector upper/lower split
    det_in_out: Bool
        Perform a detector in/out split
    wrap: Bool or str
        If True, the flags with the det splits will be wrapped to aman.det_flags. If a string, the flags with the det splits will be wrapped to aman.string

    Returns
    -------
    fm: FlagManager with the requested flags
    """

    fm = FlagManager.for_tod(aman)

    if det_left_right or det_in_out:
        xi = aman.focal_plane.xi
        xi_median = np.median(xi)    
    if det_upper_lower or det_in_out:
        eta = aman.focal_plane.eta
        eta_median = np.median(eta)
    if det_left_right:
        mask = xi <= xi_median
        fm.wrap_dets('det_left', np.logical_not(mask))
        mask = xi > xi_median
        fm.wrap_dets('det_right', np.logical_not(mask))
    if det_upper_lower:
        mask = eta <= eta_median
        fm.wrap_dets('det_lower', np.logical_not(mask))
        mask = eta > eta_median
        fm.wrap_dets('det_upper', np.logical_not(mask))
    if det_in_out:
        xi_center = np.min(xi) + 0.5 * (np.max(xi) - np.min(xi))
        eta_center = np.min(eta) + 0.5 * (np.max(eta) - np.min(eta))
        radii = np.sqrt((xi_center-xi)**2 + (eta_center-eta)**2)
        radius_median = np.median(radii)
        mask = radii <= radius_median
        fm.wrap_dets('det_in', np.logical_not(mask))
        mask = radii > radius_median
        fm.wrap_dets('det_out', np.logical_not(mask))

    
    if wrap == True:
        if 'det_flags' in aman._fields:
            aman.move('det_flags', None)
        aman.wrap('det_flags', fm)
    elif isinstance(wrap, str):
        if wrap in aman._fields:
            aman.move(wrap, None)
        aman.wrap(wrap, fm)
    return fm

def get_split_flags(aman, proc_aman=None, split_cfg=None):
    '''
    Function returning flags used for null splits consumed by the mapmaking
    and bundling codes. Fields labeled ``field_name_flag`` contain boolean
    masks and ``_avg`` are the mean of the numerical based split flags to 
    be used for observation level splits.

    Arguments
    ---------
    aman: AxisManager
        Main axis manager containing signal.
    proc_aman: AxisManager
        Preprocess axis manager, usually loaded in ``aman.preprocess``.
    split_cfg: dict
        Dictionary containing the thresholds used for cutting

    Returns
    -------
    split_aman: AxisManager
        Axis manager containing splitting flags.
        ``cuts`` field is a FlagManager containing the detector and subscan based splits used in the mapmaker.
        ``<split_name>_threshold`` fields contain the threshold used for the split.
        Other fields conatain info for obs-level splits.
    '''
    # Set default set of splits
    default_cfg = {'high_gain': 0.115, 'high_tau': 1.5e-3,
                   'det_A': 'A', 'pol_angle': 35, 'crossover': 'BL', 'right_focal_plane': 0,
                   'top_focal_plane': 0, 'central_pixels': 0.071 }
    if split_cfg is None:
        split_cfg = default_cfg

    split_aman = AxisManager(aman.dets)
    fm = FlagManager.for_tod(aman)
    # If provided split config doesn't include all of the splits in default
    for k in default_cfg.keys():
        if not k in split_cfg:
            split_cfg[k] = default_cfg[k]

    # Gain split
    fm.wrap_dets('high_gain', aman.det_cal.phase_to_pW > split_cfg['high_gain'])
    fm.wrap_dets('low_gain', aman.det_cal.phase_to_pW <= split_cfg['high_gain'])
    split_aman.wrap('gain_avg', np.nanmean(aman.det_cal.phase_to_pW))
    # Time constant split
    fm.wrap_dets('high_tau', aman.det_cal.tau_eff > split_cfg['high_tau'])
    fm.wrap_dets('low_tau', aman.det_cal.tau_eff <= split_cfg['high_tau'])
    split_aman.wrap('tau_avg', np.nanmean(aman.det_cal.tau_eff))
    # detAB split
    fm.wrap_dets('det_A', aman.det_info.wafer.pol <= split_cfg['det_A'])
    fm.wrap_dets('det_B', aman.det_info.wafer.pol > split_cfg['det_A'])
    # def pol split
    fm.wrap_dets('high_pol_angle', aman.det_info.wafer.angle > split_cfg['pol_angle'])
    fm.wrap_dets('low_pol_angle', aman.det_info.wafer.angle <= split_cfg['pol_angle'])
    # crossover split, B and L are cross over, T and R are cross under
    fm.wrap_dets('crossover', [d in split_cfg['crossover'] for d in aman.det_info.wafer.crossover])
    fm.wrap_dets('crossunder', [d not in split_cfg['crossover'] for d in aman.det_info.wafer.crossover])
    # Right/left focal plane split
    fm.wrap_dets('det_right', aman.focal_plane.xi > split_cfg['right_focal_plane'])
    fm.wrap_dets('det_left', aman.focal_plane.xi <= split_cfg['right_focal_plane'])
    # Top/bottom focal plane split
    fm.wrap_dets('det_upper', aman.focal_plane.eta > split_cfg['top_focal_plane'])
    fm.wrap_dets('det_lower', aman.focal_plane.eta <= split_cfg['top_focal_plane'])
    # Inner/outter pixel split
    r =  np.sqrt(aman.focal_plane.xi**2 + aman.focal_plane.eta**2)
    fm.wrap_dets('det_in', r < split_cfg['central_pixels'])
    fm.wrap_dets('det_out', r >= split_cfg['central_pixels'])

    # Preproc dependent splits
    if proc_aman is None:
        try:
            proc_aman = aman.preprocess
        except:
            print('Preprocess information not present, cannot generate preprocess dependent splits.')
            for k in split_cfg.keys():
                split_aman.wrap(f'{k}_threshold', split_cfg[k])
            split_aman.wrap('cuts', fm)
            return split_aman

    # This one is a bit funky to be forcing it to be noiseQ_fit, and units matter!
    if 'noiseQ_fit' in proc_aman:
        # Noise split
        if not 'high_noise' in split_cfg:
            split_cfg['high_noise'] = 3.5e-5
        fm.wrap_dets('high_noise', proc_aman.noiseQ_fit.fit[:,1] > split_cfg['high_noise'])
        fm.wrap_dets('low_noise', proc_aman.noiseQ_fit.fit[:,1] <= split_cfg['high_noise'])
        split_aman.wrap('noise_avg', np.nanmean(proc_aman.noiseQ_fit.fit[:,1]))

    if 't2p' in proc_aman:
        # T2P Leakage split
        if not 'high_leakage' in split_cfg:
            split_cfg['high_leakage'] = 1e-3
        fm.wrap_dets('high_leakage', np.sqrt(proc_aman.t2p.lamQ**2 + proc_aman.t2p.lamU**2) > split_cfg['high_leakage'])
        fm.wrap_dets('low_leakage', np.sqrt(proc_aman.t2p.lamQ**2 + proc_aman.t2p.lamU**2) <= split_cfg['high_leakage'])
        split_aman.wrap('leakage_avg', np.nanmean(np.sqrt(proc_aman.t2p.lamQ**2 + proc_aman.t2p.lamU**2)),
                    [(0, 'dets')])
    if 'hwpss_stats' in proc_aman:
        # High 2f amplitude split
        if not 'high_2f' in split_cfg:
            split_cfg['high_2f'] = 1.5e-3
        a2 = aman.det_cal.phase_to_pW*np.sqrt(proc_aman.hwpss_stats.coeffs[:,2]**2 + proc_aman.hwpss_stats.coeffs[:,3]**2)
        fm.wrap_dets('high_2f', a2 > split_cfg['high_2f'])
        fm.wrap_dets('low_2f', a2 <= split_cfg['high_2f'])
        split_aman.wrap('2f_avg', np.nanmean(a2), [(0, 'dets')])
    # Left/right subscans
    if 'turnaround_flags' in proc_aman:
        fm.wrap('scan_left', proc_aman.turnaround_flags.left_scan)
        fm.wrap('scan_right', proc_aman.turnaround_flags.right_scan)

    for k in split_cfg.keys():
        split_aman.wrap(f'{k}_threshold', split_cfg[k])

    split_aman.wrap('cuts', fm)

    return split_aman
